import re
import sys

DEVICE_TOKEN_FORMAT = re.compile(r'^[0-9a-fA-F]{64}$')
UUID_FORMAT = re.compile(
    r'^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}'
    r'-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$')
SMS_SENDER_FORMAT = re.compile(r'^[0-9]*$')
SMS_MSISDN_FORMAT = re.compile(r'^[0-9]*$')

# Python coarse version differentiation
PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

# Set version string type
if PY3:
    string_type = str
elif PY2:
    string_type = basestring

# Value selectors; device IDs, aliases, tags, etc.


def ios_channel(uuid):
    """Select a single iOS Channel"""
    if not UUID_FORMAT.match(uuid):
        raise ValueError('Invalid iOS Channel')
    return {'ios_channel': uuid.lower().strip()}


def android_channel(uuid):
    """Select a single Android Channel"""
    if not UUID_FORMAT.match(uuid):
        raise ValueError('Invalid Android Channel')
    return {'android_channel': uuid.lower().strip()}


def amazon_channel(uuid):
    """Select a single Amazon Channel"""
    if not UUID_FORMAT.match(uuid):
        raise ValueError('Invalid Amazon Channel')
    return {'amazon_channel': uuid.lower().strip()}


def device_token(token):
    """Select a single iOS device token"""
    # Ensure the device token is valid
    if not DEVICE_TOKEN_FORMAT.match(token):
        raise ValueError('Invalid device token')
    return {'device_token': token.upper().strip()}


def apid(uuid):
    """Select a single Android APID"""
    if not UUID_FORMAT.match(uuid):
        raise ValueError('Invalid APID')
    return {'apid': uuid.lower().strip()}


def channel(uuid):
    """Select a single Web Channel"""
    if not UUID_FORMAT.match(uuid):
        raise ValueError('Invalid Channel')
    return {'channel': uuid.lower().strip()}


def open_channel(uuid):
    """Select a single Open Channel"""
    if not UUID_FORMAT.match(uuid):
        raise ValueError('Invalid Open Channel')
    return {'open_channel': uuid.lower().strip()}


def sms_sender(sender):
    if not (isinstance(sender, string_type) or SMS_SENDER_FORMAT.match(sender)):
        raise ValueError('sms_sender value must be a numeric string.')
    return {'sms_sender': sender}


def sms_id(msisdn, sender):
    if not (isinstance(msisdn, string_type) or SMS_MSISDN_FORMAT.match(msisdn)):
        raise ValueError('msisdn value must be a numeric string.')
    if not (isinstance(sender, string_type) or SMS_SENDER_FORMAT.match(sender)):
        raise ValueError('sender value must be a numeric string.')
    return {'sms_id': {'sender': sender, 'msisdn': msisdn}}


def wns(uuid):
    """Select a single Windows 8 APID"""
    if not UUID_FORMAT.match(uuid):
        raise ValueError('Invalid wns')
    return {'wns': uuid.lower().strip()}


def tag(tag):
    """Select a single tag."""
    return {'tag': tag}


def alias(alias):
    """Select a single alias."""
    return {'alias': alias}


def segment(segment):
    """Select a single segment."""
    return {'segment': segment}


# Compound selectors


def or_(*children):
    """Select devices that match at least one of the given selectors.

    >>> or_(tag('sports'), tag('business'))
    {'or': [{'tag': 'sports'}, {'tag': 'business'}]}

    """
    return {'or': [child for child in children]}


def and_(*children):
    """Select devices that match all of the given selectors.

    >>> and_(tag('sports'), tag('business'))
    {'and': [{'tag': 'sports'}, {'tag': 'business'}]}

    """
    return {'and': [child for child in children]}


def not_(child):
    """Select devices that does not match the given selectors.

    >>> not_(and_(tag('sports'), tag('business')))
    {'not': {'and': [{'tag': 'sports'}, {'tag': 'business'}]}}

    """
    return {'not': child}


# Location selectors

def location(date=None, **kwargs):
    """Select a location expression.

    Location selectors are made up of either an id or an alias and a date
    period specifier. Use a date specification function to generate the time
    period specifier.

    ID location example:

    >>> from pprint import pprint
    >>> l = location(id='4oFkxX7RcUdirjtaenEQIV', date=recent_date(days=4))
    >>> pprint(l, width=76)
    {'location': {'date': {'recent': {'days': 4}},
                  'id': '4oFkxX7RcUdirjtaenEQIV'}}

    Alias location example:

    >>> l = location(us_zip='94103', date=absolute_date(
    ...    resolution='days', start='2012-01-01', end='2012-01-15'))
    >>> pprint(l, width=76)
    {'location': {'date': {'days': {'end': '2012-01-15',
                                    'start': '2012-01-01'}},
                  'us_zip': '94103'}}

    """
    if not len(kwargs) == 1:
        raise ValueError('Must specify a single location id or alias')
    if date is None:
        raise ValueError('Must specify a time period specifier')
    kwargs['date'] = date
    return {'location': kwargs}


def recent_date(**kwargs):
    """Select a recent date range for a location selector.

    :keyword resolution: One keyword time resolution specifier, e.g. ``hours``
        or ``days``.
    :type resolution: int

    >>> recent_date(months=6)
    {'recent': {'months': 6}}
    >>> recent_date(weeks=3)
    {'recent': {'weeks': 3}}
    """
    if not len(kwargs) == 1:
        raise ValueError('Must specify a single date resolution')
    resolution = list(kwargs.keys())[0]
    value = list(kwargs.values())[0]

    if resolution not in ('minutes' 'hours' 'days' 'weeks' 'months' 'years'):
        raise ValueError('Invalid date resolution: %s' % resolution)
    payload = {'recent': {resolution: value}}
    return payload


def absolute_date(resolution, start, end):
    """Select an absolute date range for a location selector.

    :keyword resolution: Time resolution specifier, e.g. ``hours`` or ``days``.
    :keyword start: UTC start time in ISO 8601 format.
    :keyword end: UTC end time in ISO 8601 format.

    >>> from pprint import pprint
    >>> d = absolute_date(resolution='months', start='2013-01', end='2013-06')
    >>> pprint(d)
    {'months': {'end': '2013-06', 'start': '2013-01'}}
    >>> d = absolute_date(resolution='minutes', start='2012-01-01 12:00',
    ...         end='2012-01-01 12:45')
    >>> pprint(d, width=76)
    {'minutes': {'end': '2012-01-01 12:45', 'start': '2012-01-01 12:00'}}

    """
    if resolution not in ('minutes' 'hours' 'days' 'weeks' 'months' 'years'):
        raise ValueError('Invalid date resolution: %s' % resolution)

    payload = {resolution: {'start': start, 'end': end}}
    return payload


def named_user(name):
    return {'named_user': name}



from urbanairship import common
from urbanairship.push import ScheduledPush


class ScheduledList(common.IteratorParent):
    """
    Iterator for listing all scheduled messages.

    :ivar limit: Number of entries to fetch in a paginated request.
    :returns Each ``next`` returns a :py:class:`ScheduledPush` object.
    """
    next_url = None
    data_attribute = 'schedules'
    id_key = 'url'
    instance_class = ScheduledPush

    def __init__(self, airship, limit=None):
        self.next_url = airship.urls.get('schedules_url')
        params = {'limit': limit} if limit else {}
        super(ScheduledList, self).__init__(airship, params)


def scheduled_time(timestamp):
    """Specify a time for the delivery of this push.

    :param timestamp: A ``datetime.datetime`` object.

    """

    return {'scheduled_time': timestamp.strftime('%Y-%m-%dT%H:%M:%S')}


def local_scheduled_time(timestamp):
    """Specify a time for the delivery of this push in device local time.

    :param timestamp: A ``datetime.datetime`` object.

    """

    return {'local_scheduled_time': timestamp.strftime('%Y-%m-%dT%H:%M:%S')}


def best_time(timestamp):
    """Specify a date to send the push at the best time per-device.
    Only YYYY_MM_DD are needed. Hour/minute/second information is discarded.

    :param timestamp: A ``datetime.datetime object.
    """

    return {'best_time': {'send_date': timestamp.strftime('%Y-%m-%d')}}



import json
import logging

from urbanairship.devices import ChannelTags
from urbanairship import common

logger = logging.getLogger('urbanairship')


class NamedUser(object):
    """Perform various operations on a named user object"""

    def __init__(self, airship, named_user_id=None):

        self._airship = airship
        self.named_user_id = named_user_id

    def associate(self, channel_id, device_type):
        """Associate a channel with a named user ID

        :param channel_id: The ID of the channel you would like to associate
            with the named user
        :param device_type: The device type of the channel
        :return:
        """
        if not self.named_user_id:
            raise ValueError('named_user_id is required for association')

        body = json.dumps(
            {
                'channel_id': channel_id,
                'device_type': device_type,
                'named_user_id': self.named_user_id
            }
        ).encode('utf-8')
        response = self._airship._request(
            'POST',
            body,
            self._airship.urls.get('named_user_associate_url'),
            'application/json',
            version=3
        )
        return response

    def disassociate(self, channel_id, device_type):
        """Disassociate a channel with a named user ID

        :param channel_id: The ID of the channel you would like to disassociate
        :param device_type: The device type of the channel
        :return:
        """

        payload = {'channel_id': channel_id, 'device_type': device_type}

        if self.named_user_id:
            payload['named_user_id'] = self.named_user_id

        body = json.dumps(payload).encode('utf-8')
        response = self._airship._request(
            'POST',
            body,
            self._airship.urls.get('named_user_disassociate_url'),
            'application/json',
            version=3
        )

        return response

    def lookup(self):
        """Lookup a single named user

        :return: The named user payload for the named user ID
        """
        response = self._airship._request(
            'GET',
            None,
            self._airship.urls.get('named_user_url'),
            'application/json',
            version=3,
            params={'id': self.named_user_id}
        )
        return response.json()

    def tag(self, group, add=None, remove=None, set=None):
        """Add, remove, or set tags on a named user
        :param add: A list of tags to add
        :param remove: A list of tags to remove
        :param set: A list of tags to set
        :param group: The Tag group for the add, remove, and set operations
        """
        if self.named_user_id:
            payload = {'audience': {'named_user_id': self.named_user_id}}
        else:
            raise ValueError('A named user ID is required for modifying tags')

        if add:
            if set:
                raise ValueError('A tag request can only contain an add or '
                                 'remove field, both, or a single set field')
            payload['add'] = {group: add}

        if remove:
            if set:
                raise ValueError('A tag request can only contain an add or '
                                 'remove field, both, or a single set field')
            payload['remove'] = {group: remove}

        if set:
            payload['set'] = {group: set}
        if not add and not remove and not set:
            raise ValueError('An add, remove, or set field was not set')

        body = json.dumps(payload).encode('utf-8')
        response = self._airship._request(
            'POST',
            body,
            self._airship.urls.get('named_user_tag_url'),
            'application/json',
            version=3
        )

        return response.json()

    @classmethod
    def from_payload(cls, payload):
        """
        Create NamedUser object based on results from a NamedUserList iterator.
        :param payload: Payload used to create the NamedUser object

        """
        for key in payload:
            setattr(cls, key, payload[key])

        return cls


class NamedUserList(common.IteratorParent):
    """Retrieves a list of NamedUsers"""
    next_url = None
    data_attribute = 'named_users'

    def __init__(self, airship):
        self.next_url = airship.urls.get('named_user_url')
        super(NamedUserList, self).__init__(airship, None)


class NamedUserTags(ChannelTags):
    """Modify the tags for named users"""

    def __init__(self, airship):
        super(NamedUserTags, self).__init__(airship)
        self.url = airship.urls.get('named_user_tag_url')

    def set_audience(self, user_ids):
        self.audience['named_user_id'] = user_ids



import logging

logger = logging.getLogger('urbanairship')


class LocationFinder(object):
    def __init__(self, airship):
        self.airship = airship

    def name_lookup(self, name, location_type=None):
        """Lookup a location by name

        :param name: Name of the location to look up
        :param location_type: Location type
        :return: Information about the location
        """

        params = {'q': name}
        url = self.airship.urls.get('location_url')

        if location_type:
            params['type'] = location_type

        resp = self.airship._request(
            'GET',
            None,
            url,
            version=3,
            params=params
        )
        logger.info(u'Retrieved location info by name: %s' % name)
        return resp.json()

    def coordinates_lookup(self, latitude, longitude, location_type=None):
        """Lookup a location by coordinates

        :param latitude: Latitude of the location
        :param longitude: Longitude of the location
        :param location_type: Type of the location (optional)
        :return: Information about the location
        """

        coor_are_valid = isinstance(latitude, (int, float)) \
            and isinstance(longitude, (int, float))

        if not coor_are_valid:
            raise TypeError('latitude and longitude need to be numbers')

        url = '{base_url}{latitude},{longitude}'.format(
            base_url=self.airship.urls.get('location_url'),
            latitude=str(latitude),
            longitude=str(longitude))

        params = {}

        if location_type:
            params['type'] = location_type

        resp = self.airship._request(
            'GET',
            None,
            url,
            version=3,
            params=params
        )
        logger.info('Retrieved location info by coordinates: %s,%s' % (
            str(latitude), str(longitude)))
        return resp.json()

    def bounding_box_lookup(
            self, lat1, long1, lat2, long2, location_type=None
    ):
        """Lookup a location by bounding box

        :param lat1: First latitude of the location
        :param long1: First longitude of the location
        :param lat2: Second latitude of the location
        :param long2: Second longitude of the location
        :param location_type: Type of the location (optional)
        :return: Information about the location
        """

        loc1_valid = isinstance(lat1, (float, int)) \
            and isinstance(long1, (float, int))
        loc2_valid = isinstance(lat2, (float, int)) \
            and isinstance(long2, (float, int))

        if not loc1_valid or not loc2_valid:
            raise TypeError('lat1, long1, lat2, and long2 need to be numbers')

        url = '{base_url}{lat1},{long1},{lat2},{long2}'.format(
            base_url=self.airship.urls.get('location_url'),
            lat1=str(lat1),
            long1=str(long1),
            lat2=str(lat2),
            long2=str(long2)
        )
        params = {}

        if location_type:
            params['type'] = location_type

        resp = self.airship._request(
            'GET',
            None,
            url,
            version=3,
            params=params
        )
        logger.info(
            'Retrieved location info by bounding box: %s,%s,%s,%s' % (
                str(lat1), str(long1), str(lat2), str(long2)))
        return resp.json()

    def alias_lookup(self, from_alias):
        """Lookup a location by alias or list of aliases
        :param from_alias: alias of the location
        :return: Information about the location
        """

        params = {}

        if isinstance(from_alias, list):
            for alias in from_alias:
                alias_parts = alias.split('=')
                params[alias_parts[0]] = alias_parts[1]
        else:
            alias_parts = from_alias.split('=')
            params[alias_parts[0]] = alias_parts[1]

        resp = self.airship._request(
            'GET',
            None,
            self.airship.urls.get('location_url') + 'from-alias',
            version=3,
            params=params
        )
        logger.info('Retrieved location info by alias: %s' % from_alias)
        return resp.json()

    def polygon_lookup(self, polygon_id, zoom):
        """ Lookup the location by polygon id
        :param polygon_id: polygon id of the location
        :param zoom: zoom level of the desired location information
        :return: Information about the location
        """

        if not isinstance(zoom, int) or zoom < 1 or zoom > 20:
            raise TypeError('zoom needs to be an integer between 1 and 20')
        params = {'zoom': str(zoom)}
        resp = self.airship._request(
            'GET',
            None,
            self.airship.urls.get('location_url') + polygon_id,
            version=3,
            params=params
        )
        logger.info(
            'Retrieved location info by polygon id: %s and zoom: %s' % (
                polygon_id, str(zoom)
            )
        )
        return resp.json()

    def date_ranges(self):
        resp = self.airship._request(
            'GET',
            None,
            self.airship.urls.get('segments_url') + 'dates/',
            version=3,
        )
        return resp.json()



import json
import logging

logger = logging.getLogger('urbanairship')


class ChannelUninstall(object):
    _airship = None

    def __init__(self, airship):
        self._airship = airship

    def uninstall(self, channels):
        chan_num = len(channels)

        if chan_num > 200:
            raise ValueError(
                ('Maximum of 200 channel uninstalls exceeded. '
                 '({0} channels)').format(chan_num)
            )

        body = json.dumps(channels)
        url = self._airship.urls.get('channel_url') + 'uninstall/'

        response = self._airship._request('POST', body, url, version=3)
        logger.info('Successfully uninstalled {0} channels'.format(chan_num))

        return response



import logging
import datetime
import six

logger = logging.getLogger('urbanairship')


class Unauthorized(Exception):
    """Raised when we get a 401 from the server"""


class AirshipFailure(Exception):
    """Raised when we get an error response from the server.

    :param args: For backwards compatibility, ``*args`` includes the status and
        response body.

    """

    error = None
    error_code = None
    details = None
    response = None

    def __init__(self, error, error_code, details, response, *args):
        self.error = error
        self.error_code = error_code
        self.details = details
        self.response = response
        super(AirshipFailure, self).__init__(*args)

    @classmethod
    def from_response(cls, response):
        """
        Instantiate a ValidationFailure from a Response object
        :param response: response object used to create failure obj
        """
        try:
            payload = response.json()
            error = payload.get('error')
            error_code = payload.get('error_code')
            details = payload.get('details')
        except (ValueError, TypeError, KeyError):
            error = response.reason
            error_code = response.status_code
            details = response.content

        logger.error(
            "Request failed with status %d: '%s %s': %s",
            response.status_code, error_code, error, details)

        return cls(
            error,
            error_code,
            details,
            response,
            response.status_code,
            response.content
        )


@six.python_2_unicode_compatible
class IteratorDataObj(object):
    @classmethod
    def from_payload(cls, payload, device_key=None, airship=None):
        obj = cls()
        if device_key:
            obj.device_type = device_key
        if device_key and payload[device_key]:
            obj.id = payload[device_key]
        if airship:
            obj.airship = airship
        for key in payload:
            try:
                val = datetime.datetime.strptime(
                    payload[key],
                    '%Y-%m-%d %H:%M:%S'
                )
            except (TypeError, ValueError):
                val = payload[key]
            setattr(obj, key, val)
        return obj

    def __str__(self):
        print_str = ""
        for attr in dir(self):
            if(
                not attr.startswith('__') and
                not hasattr(getattr(self, attr), '__call__')
            ):
                print_str += attr + ': ' + str(getattr(self, attr)) + ', '
        return print_str[:-2]


class IteratorParent(six.Iterator):
    next_url = None
    data_attribute = None
    data_list = None
    params = None
    id_key = None
    instance_class = IteratorDataObj

    def __init__(self, airship, params):
        self.airship = airship
        self.params = params
        self._token_iter = iter(())

    def __iter__(self):
        return self

    def __next__(self):
        try:
            return self.instance_class.from_payload(
                next(self._token_iter),
                self.id_key,
                self.airship
            )
        except StopIteration:
            if self._load_page():
                return self.instance_class.from_payload(
                    next(self._token_iter),
                    self.id_key,
                    self.airship
                )
            else:
                raise StopIteration

    def _load_page(self):
        if not self.next_url:
            return False
        response = self.airship.request(
            method='GET',
            body=None,
            url=self.next_url,
            version=3,
            params=self.params
        )
        self.params = None
        self._page = response.json()
        check_url = self._page.get('next_page')
        if check_url == self.next_url:
            return False
        self.next_url = check_url
        self._token_iter = iter(self._page[self.data_attribute])
        return True



import json
import logging
import re

logger = logging.getLogger('urbanairship')

VALID_EMAIL = re.compile(r'(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)')
VALID_ISO_8601 = re.compile(
    '^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-(3[01]|0[1-9]|[12][0-9])'
    'T(2[0-3]|[01][0-9]):([0-5][0-9]):([0-5][0-9])(\\.[0-9]+)?(Z)?$'
    )


class Email(object):
    """Register and uninstall an Email object.

    Please see the email documentation for important information about
    opt-in times and email types.
    https://docs.urbanairship.com/api/ua/#operation/api/channels/email

    :param address: Required. The email address the object represents.
    :param commercial_opted_in: Optional. A string in ISO 8601 format that
        represents the time explicit permission was received from the user
        to accept commercial emails.
    :param commercial_opted_out: Optional. A string in ISO 8601 format that
        represents the time a user opted out of commercial emails.
    :param transactional_opted_in: Optional. A string in ISO 8601 format
        that represents the time a user explicitly opted in to transactional
        emails.
    :param transactional_opted_out: Optional. A string in ISO 8601 formation
        that represents the time a user explicitly opted out of transactional
        emails.
    :param locale_country: Optional. The device property tag related
        to locale country setting.
    :param locale_language: Optional. The device property tag related
        to locale language setting.
    :param timezone: Optional. The deice property tag related to
        timezone setting.
    :param template_fields: For use with CreateAndSend with inline templates.
        A dict of template field names and their substitution values.
    """

    def __init__(self, airship, address, commercial_opted_in=None,
                 commercial_opted_out=None, transactional_opted_in=None,
                 transactional_opted_out=None, locale_country=None,
                 locale_language=None, timezone=None, template_fields=None):
        self.airship = airship
        self.address = address
        self.commercial_opted_in = commercial_opted_in
        self.commercial_opted_out = commercial_opted_out
        self.transactional_opted_in = transactional_opted_in
        self.transactional_opted_out = transactional_opted_out
        self.locale_country = locale_country
        self.locale_language = locale_language
        self.timezone = timezone
        self.template_fields = template_fields
        self._email_type = 'email'  # only acceptable value at this time
        self.channel_id = None

    @property
    def template_fields(self):
        return self._template_fields

    @template_fields.setter
    def template_fields(self, value):
        if not isinstance(value, (dict, type(None))):
            raise TypeError('template_fields must be a dict')

        self._template_fields = value

    @property
    def address(self):
        return self._address

    @address.setter
    def address(self, value):
        if not VALID_EMAIL.match(value) and value is not None:
            raise ValueError('Invalid email address')
        self._address = value

    @property
    def commercial_opted_in(self):
        return self._commercial_opted_in

    @commercial_opted_in.setter
    def commercial_opted_in(self, value):
        if value is not None and not VALID_ISO_8601.match(value):
            raise ValueError('Must use ISO 8601 timestamp format')
        self._commercial_opted_in = value

    @property
    def commercial_opted_out(self):
        return self._commercial_opted_out

    @commercial_opted_out.setter
    def commercial_opted_out(self, value):
        if value is not None and not VALID_ISO_8601.match(value):
            raise ValueError('Must use ISO 8601 timestamp format')
        self._commercial_opted_out = value

    @property
    def transactional_opted_in(self):
        return self._transactional_opted_in

    @transactional_opted_in.setter
    def transactional_opted_in(self, value):
        if value is not None and not VALID_ISO_8601.match(value):
            raise ValueError('Must use ISO 8601 timestamp format')
        self._transactional_opted_in = value

    @property
    def transactional_opted_out(self):
        return self._transactional_opted_out

    @transactional_opted_out.setter
    def transactional_opted_out(self, value):
        if value is not None and not VALID_ISO_8601.match(value):
            raise ValueError('Must use ISO 8601 timestamp format')
        self._transactional_opted_out = value

    @property
    def create_and_send_audience(self):
        audience = {
            'ua_address': self.address,
        }
        if self.commercial_opted_in:
            audience['ua_commercial_opted_in'] = self.commercial_opted_in
        if self.transactional_opted_in:
            audience['ua_transactional_opted_in'] = self.transactional_opted_in
        if self.template_fields:
            audience.update(self.template_fields)

        return audience

    def register(self):
        """Create a new email channel or unsubscribe an existing email
        channel from receiving commercial emails.
        To unsubscribe an existing channel, set email_opt_in_level
        to none.

        :return: The response object from the API.
        """
        url = self.airship.urls.get('email_url')
        reg_payload = {
            'channel': {
                'type': self._email_type,
                'address': self.address,
            }
        }

        if self.commercial_opted_in:
            reg_payload['channel']['commercial_opted_in'] = \
                self.commercial_opted_in
        if self.commercial_opted_out:
            reg_payload['channel']['commercial_opted_out'] = \
                self.commercial_opted_out
        if self.transactional_opted_in:
            reg_payload['channel']['transactional_opted_in'] = \
                self.transactional_opted_in
        if self.transactional_opted_out:
            reg_payload['channel']['transactional_opted_out'] = \
                self.transactional_opted_out

        if self.locale_language is not None:
            reg_payload['channel']['locale_language'] = self.locale_language
        if self.locale_country is not None:
            reg_payload['channel']['locale_country'] = self.locale_country
        if self.timezone is not None:
            reg_payload['channel']['timezone'] = self.timezone

        body = json.dumps(reg_payload).encode('utf-8')

        response = self.airship.request(
            method='POST',
            body=body,
            url=url,
            version=3
        )

        if response.status_code == 201:
            self.channel_id = response.json().get('channel_id')
            logger.info(
                'Successfully created channel with channel_id %s' % (
                    self.channel_id
                )
            )
        elif response.status_code == 200:
            self.channel_id = response.json().get('channel_id')
            logger.info(
                'Successful registration call made to channel_id %s' % (
                    self.channel_id
                )
            )

        return response

    def uninstall(self):
        """Removes an email address from Urban Airship. Use with caution.
        If the uninstalled email address opts-in again, it will generate
        a new channel_id.
        The new channel_id cannot be reassociated with any opt-in
        information, tags, named users, insight reports, or other
        information from the uninstalled email channel.

        :return: The response object from the API"""

        url = self.airship.urls.get('email_uninstall_url')
        uninstall_payload = {'email_address': self.address}

        body = json.dumps(uninstall_payload).encode('utf-8')

        response = self.airship.request(
            method='POST',
            body=body,
            url=url,
            version=3
        )

        logger.info('Uninstalled email address: %s' % self.address)

        return response


class EmailTags(object):
    """Add, remove or set tags for a list of email addresses

    :param address: an email address to mutate tags for
    """
    def __init__(self, airship, address):
        self.airship = airship
        self.url = airship.urls.get('email_tags_url')
        self.address = address
        self.add_group = {}
        self.remove_group = {}
        self.set_group = {}
        self._payload = {}

    @property
    def address(self):
        return self._address

    @address.setter
    def address(self, value):
        if not VALID_EMAIL.match(value):
            raise ValueError('addresses must be and email address')
        self._address = value

    @property
    def tags(self):
        return self._tags

    @tags.setter
    def tags(self, value):
        if not isinstance(value, list):
            raise ValueError('tags must be input as a list')
        self._tags = value

    def add(self, group, tags):
        """
        add tags to a given tag group
        :param group: the tag group to add to
        :param tags: a list of tags to add
        """
        self.add_group[group] = tags

    def remove(self, group, tags):
        """
        remove tags from a given tag group
        :param group: the tag group to remove tags from
        :param tags: a list of tags to remove
        """
        self.remove_group[group] = tags

    def set(self, group, tags):
        """
        replace all tags on the given tag group with these tags
        :param group: the tag group to set tags on
        :param tags: a list of tags to set
        """
        self.set_group[group] = tags

    def send(self):
        """
        commit add, remove and set operations. set cannot be used with
        add and remove.
        :return: the response object from the api
        """
        if not self.add_group and not self.remove_group and not self.set_group:
            raise ValueError('at least one add, remove or set group must exist')
        self._payload['audience'] = {'email_address': self.address}

        if self.set_group:
            if self.add_group or self.remove_group:
                raise ValueError('set cannot be used with remove or add groups')
            self._payload['set'] = self.set_group

        if self.add_group:
            self._payload['add'] = self.add_group

        if self.remove_group:
            self._payload['remove'] = self.remove_group

        body = json.dumps(self._payload).encode('utf-8')

        response = self.airship.request(
            method='POST',
            body=body,
            url=self.url,
            version=3
        )

        return response



import datetime
import logging
from urbanairship import common

logger = logging.getLogger('urbanairship')


class ChannelInfo(object):
    """Information object for iOS, Android, Amazon, web, and open channels.

    :ivar address: Replaces ``push_address`` for open channels.
    :ivar alias: Alias associated with this device, if any.
    :ivar background: Bool; whether the device is opted in to background push.
    :ivar channel_id: Channel ID for the device.
    :ivar created: UTC datetime when the system initially saw the device.
    :ivar device_type: Type of the device, e.g. ``ios``.
    :ivar installed: Bool; whether the app is installed on the device.
    :ivar last_registration: UTC datetime when the system last received a
        registration call for the device.
    :ivar named_user_id: Named user associated with this device, if any.
    :ivar opt_in: Bool; whether the device is opted in to push or other visible
        notifications.
    :ivar push_address: Address we use to push to the device (device token,
        GCM registration ID, etc,). Not present for open channels (see
        ``address`` above).
    :ivar tag_groups: Tags associated with non-"device" tag groups, if any.
    :ivar tags: List of tags associated with this device, if any.
    :ivar ios: iOS specific information, e.g. ``badge`` and ``quiet_time``.
    :ivar open: Open channel specific information, e.g. ``identifiers`` and
        ``open_platform_name``.
    :ivar web: Web notify specific information, e.g. ``subscription``.

    """

    airship = None
    address = None
    alias = None
    background = None
    channel_id = None
    created = None
    device_type = None
    installed = None
    last_registration = None
    opt_in = None
    push_address = None
    tag_groups = None
    tags = None
    ios = None
    open = None
    web = None

    def __init__(self, airship):
        self.airship = airship

    @classmethod
    def from_payload(cls, payload, device_key, airship):
        """Create based on results from a ChannelList iterator."""
        obj = cls(airship)
        obj.channel_id = payload[device_key]
        if airship:
            obj.airship = airship
        for key in payload:
            if key in ('created', 'last_registration'):
                try:
                    payload[key] = datetime.datetime.strptime(
                        payload[key], '%Y-%m-%dT%H:%M:%S'
                    )
                except:
                    payload[key] = 'UNKNOWN'
            setattr(obj, key, payload[key])
        return obj

    def lookup(self, channel_id):
        """Fetch metadata from a channel ID"""
        start_url = self.airship.urls.get('channel_url')
        data_attribute = 'channel'
        id_key = 'channel_id'
        params = {}
        url = start_url + channel_id
        response = self.airship._request(
            method='GET',
            body=None,
            url=url,
            version=3,
            params=params
        )
        payload = response.json()
        return self.from_payload(payload[data_attribute], id_key, self.airship)


class DeviceInfo(object):
    """Information object for a single device.

    :ivar active: bool; Whether the device is opted in to push or other visible
        notifications.
    :ivar alias: Alias associated with this device, if any.
    :ivar created: UTC datetime when the system initially saw the device.
    :ivar device_type: Type of the device, e.g. ``device_token``, ``apid``.
    :ivar id: Device identifier. Also available at the attribute named by the
        ``device_type``.
    :ivar tags: List of tags associated with this device, if any.
    :ivar apid: Same as device identifier if apid device type.
    :ivar device_token: Same as device identifier if device_token device type.

    """

    id = None
    device_type = None
    active = None
    tags = None
    alias = None

    def __init__(self, airship):
        self.airship = airship

    @classmethod
    def from_payload(cls, payload, device_key, airship):
        """Create based on results from a DeviceTokenList or APIDList iterator.

        """
        obj = cls(airship)
        obj.id = payload[device_key]
        obj.device_type = device_key
        for key in payload:
            if key in 'created':
                try:
                    payload[key] = datetime.datetime.strptime(
                        payload[key], '%Y-%m-%d %H:%M:%S'
                    )
                except:
                    payload[key] = 'UNKNOWN'
            setattr(obj, key, payload[key])
        return obj


class DeviceTokenList(common.IteratorParent):
    """Iterator for listing all device tokens for this application.

    :ivar limit: Number of entries to fetch in each page request.
    :returns: Each ``next`` returns a :py:class:`DeviceInfo` object.

    """
    next_url = None
    data_attribute = 'device_tokens'
    id_key = 'device_token'
    instance_class = DeviceInfo

    def __init__(self, airship, limit=None):
        self.next_url = airship.urls.get('device_token_url')
        params = {'limit': limit} if limit else {}
        super(DeviceTokenList, self).__init__(airship, params)


class ChannelList(common.IteratorParent):
    """Iterator for listing all channels for this application.

    :ivar limit: Number of entries to fetch in each page request.
    :ivar start_channel: uuid representing the channel_id to start with.
    :returns: Each ``next`` returns a :py:class:`ChannelInfo` object.

    """

    next_url = None
    data_attribute = 'channels'
    id_key = 'channel_id'
    instance_class = ChannelInfo

    def __init__(self, airship, limit=None, start_channel=None):
        self.next_url = airship.urls.get('channel_url')
        channel_params = {}
        if limit:
            channel_params['limit'] = limit
        if start_channel:
            channel_params['start'] = start_channel

        super(ChannelList, self).__init__(airship, params=channel_params)


class APIDList(common.IteratorParent):
    """Iterator for listing all APIDs for this application.

    :ivar limit: Number of entries to fetch in each page request.
    :returns: Each ``next`` returns a :py:class:`DeviceInfo` object.

    """
    next_url = None
    data_attribute = 'apids'
    id_key = 'apid'
    instance_class = DeviceInfo

    def __init__(self, airship, limit=None):
        self.next_url = airship.urls.get('apid_url')
        params = {'limit': limit} if limit else {}
        super(APIDList, self).__init__(airship, params)



from .reports import (
    IndividualResponseStats,
    ResponseList,
    DevicesReport,
    OptInList,
    OptOutList,
    PushList,
    ResponseReportList,
    AppOpensList,
    TimeInAppList,
    CustomEventsList,
)



import datetime
import json
import logging
import re

logger = logging.getLogger('urbanairship')

VALID_UUID = re.compile(r'[0-9a-f]{12}4[0-9a-f]{3}[89ab][0-9a-f]{15}\Z')


class OpenChannel(object):
    channel_id = None
    address = None
    open_platform = None
    identifiers = None
    opt_in = None
    installed = None
    created = None
    last_registration = None
    tags = None
    template_fields = None

    def __init__(self, airship):
        self.airship = airship

    @property
    def create_and_send_audience(self):
        if not self.address:
            raise ValueError('open channel address must be set')

        audience = {'ua_address': self.address}

        if self.template_fields:
            audience.update(self.template_fields)

        return audience

    def create(self):
        """Create this OpenChannel object with the API."""

        if not self.address:
            raise ValueError('Must set address before creation.')

        if not self.open_platform:
            raise ValueError('Must set open_platform before creation.')

        if not isinstance(self.opt_in, bool):
            raise ValueError('Must set opt_in before creation.')

        if self.tags and not isinstance(self.tags, list):
            raise TypeError('"tags" must be a list')

        url = self.airship.urls.get('open_channel_url')

        channel_data = {
            'type': 'open',
            'address': self.address,
            'opt_in': self.opt_in,
            'open': {'open_platform_name': self.open_platform}
        }

        if self.tags:
            channel_data['tags'] = self.tags
        if self.identifiers:
            channel_data['open']['identifiers'] = self.identifiers

        body = json.dumps({'channel': channel_data})
        response = self.airship.request(
            method='POST',
            body=body,
            url=url,
            version=3
        )

        self.channel_id = response.json().get('channel_id')

        logger.info(
            'Successful open channel creation: %s (%s)',
            self.channel_id, self.address
        )

        return response

    def update(self):
        """Update this OpenChannel object."""

        if not self.address and not self.channel_id:
            raise ValueError('Must set address or channel ID to update.')

        if not self.open_platform:
            raise ValueError('Must set open_platform.')

        if not isinstance(self.opt_in, bool):
            raise ValueError('Must set opt_in.')

        if not self.address and self.opt_in is True:
            raise ValueError('Address must be set for opted in channels.')

        url = self.airship.urls.get('open_channel_url')

        channel_data = {
            'type': 'open',
            'open': {'open_platform_name': self.open_platform},
            'opt_in': self.opt_in
        }
        if self.channel_id:
            channel_data['channel_id'] = self.channel_id
        if self.address:
            channel_data['address'] = self.address
        if self.tags:
            channel_data['tags'] = self.tags
        if self.identifiers:
            channel_data['open']['identifiers'] = self.identifiers

        body = json.dumps({'channel': channel_data})
        response = self.airship.request(
            method='POST',
            body=body,
            url=url,
            version=3
        )

        self.channel_id = response.json().get('channel_id')

        logger.info(
            'Successful open channel update: %s (%s)',
            self.channel_id, self.address
        )

        return response

    @classmethod
    def from_payload(cls, payload, airship):
        """Instantiate an OpenChannel from a payload."""
        obj = cls(airship)
        for key in payload:
            # Extract the open channel data
            if key == 'open':
                obj.open_platform = payload['open'].get('open_platform_name')
                obj.identifiers = payload['open'].get('identifiers', [])
                continue

            if key in ('created', 'last_registration'):
                try:
                    payload[key] = datetime.datetime.strptime(
                        payload[key], '%Y-%m-%dT%H:%M:%S'
                    )
                except (KeyError, ValueError):
                    payload[key] = 'UNKNOWN'
            setattr(obj, key, payload[key])

        return obj

    def lookup(self, channel_id):
        """Retrieves an open channel from the provided channel ID."""
        url = self.airship.urls.get('channel_url') + channel_id
        response = self.airship._request(
            method='GET',
            body=None,
            url=url,
            version=3
        )
        payload = response.json().get('channel')

        return self.from_payload(payload, self.airship)

    def uninstall(self):
        """Mark this OpenChannel object uninstalled"""
        url = self.airship.urls.get('open_channel_url') + 'uninstall/'
        if self.address is None or self.open_platform is None:
            raise ValueError(
                '"address" and "open_platform" are required attributes'
            )

        channel_data = {
            "address": self.address,
            "open_platform_name": self.open_platform
        }

        body = json.dumps(channel_data)
        response = self.airship.request(
            method='POST',
            body=body,
            url=url,
            version=3
        )

        logger.info(
            'Successfully uninstalled open channel %s'
            % channel_data['open_platform_name']
        )

        return response



# coding=utf-8
import re
import sys
import collections
import warnings

# Python coarse version differentiation
PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

# Set version string type
if PY3:
    string_type = str
elif PY2:
    string_type = basestring

# Valid autobadge values: auto, +N, -N
VALID_AUTOBADGE = re.compile(r'^(auto|[+-][\d]+)$')
VALID_ANDROID_CATEGORIES = [
    "alarm", "call", "email", "err", "event", "msg", "promo",
    "recommendation", "service", "social", "status", "sys", "transport"
]
VALID_ICON_COLOR = re.compile(r'^#[0-9a-f]{6}$')


def notification(alert=None, ios=None, android=None, amazon=None,
                 web=None, wns=None, actions=None, interactive=None,
                 in_app=None, open_platform=None, sms=None, email=None):
    """Create a notification payload.

    :keyword alert: A simple text alert, applicable for all platforms.
    :keyword ios: An iOS platform override, as generated by :py:func:`ios`.
    :keyword android: An Android platform override, as generated by
        :py:func:`android`.
    :keyword amazon: An Amazon platform override, as generated by
        :py:func:`amazon`.
    :keyword web: A web platform override, as generated by :py:func:`web`.
    :keyword wns: A WNS platform override, as generated by :py:func:`wns`.
    :keyword sms: An Sms platform override as generated by :py:func:`sms`.
    :keyword actions: Used to perform a defined task.
    :keyword interactive: A dictionary with two attributes -- "type" and
        "button_actions", as generated by :py:func:`interactive`.
    :keyword open_platform: A dictionary with open platform name keys, and
        open platform overrides, as generated by :py:func:`open_platform`.

    """
    payload = {}
    if alert is not None:
        payload['alert'] = alert
    if actions is not None:
        payload['actions'] = actions
    if ios is not None:
        payload['ios'] = ios
    if android is not None:
        payload['android'] = android
    if amazon is not None:
        payload['amazon'] = amazon
    if web is not None:
        payload['web'] = web
    if wns is not None:
        payload['wns'] = wns
    if sms is not None:
        payload['sms'] = sms
    if email is not None:
        payload['email'] = email
    if interactive is not None:
        payload['interactive'] = interactive
    if in_app is not None:
        payload['in_app'] = in_app
    if open_platform is not None:
        for platform, overrides in open_platform.items():
            payload['open::' + platform] = overrides
    if not payload:
        raise ValueError('Notification body may not be empty')
    return payload


def ios(alert=None, badge=None, sound=None, content_available=False,
        extra=None, expiry=None, interactive=None, category=None, title=None,
        mutable_content=None, subtitle=None, media_attachment=None,
        priority=None, collapse_id=None, thread_id=None):
    """iOS/APNS specific platform override payload.

    :keyword alert: iOS format alert, as either a string or dictionary. If
        a dictionary, the accepted keys must match those in
        `Apple's Payload Key Reference
        <https://developer.apple.com/library/archive/documentation/NetworkingInternet/Conceptual/RemoteNotificationsPG/PayloadKeyReference.html>_.
        Some options require iOS 11 or 12 on the device.
    :keyword badge: An integer badge value or an *autobadge* string.
    :keyword sound: If a string, the name of a sound file in the app bundle.
        If a dict, the 'name' key specifies the name of the sound file. Some
        options require iOS 12 or above on the device.
    :keyword content_available: If true, the payload is delivered to your app
        in the background. This flag is automatically set to true if there
        is an In-App Message in the payload.
    :keyword extra: A set of key/value pairs to include in the push payload
        sent to the device.
    :keyword expiry: An integer or time set in UTC as a string
    :keyword interactive: A dictionary with two attributes -- "type" and
        "button_actions", as generated by :py:func:`interactive`.
    :keyword category: A keyword used to categorize the notification. Must be
        a string
    :keyword title: Sets the title of the notification for Apple Watch. Must
        be a string
    :keyword mutable_content: Optional, a boolean. Defaults to false.
        When set to true, content may be modified by an extension.
        This flag will be automatically set to true if there is a
        media_attachment in the payload. iOS 10 or above.
    :keyword subtitle: Optional, a string that will display below the title of
        the notification.
        This is provided as a convenience for setting the subtitle in the alert
        JSON object.
        If a subtitle is also defined in the alert JSON object, this value is
        ignored. iOS 10 or above.
    :keyword media_attachment: Optional, a dictionary object `Media Attachment
        <https://docs.urbanairship.com/api/ua/#media-attachment>`_.
        Specifies a media attachment to be handled by the UA Media Attachment
        Extension.
    :keyword collapse_id: Optional, a string. When there is a newer message
        that renders an older, related message irrelevant to the client app,
        the new message replaces the older message with the same collapse_id.
        Similar to the GCM collapse_key. The value of this key must not exceed
        64 bytes. iOS 10 or above.
    :keyword thread_id: Optional, a string. A unique identifier used to group
        notifications into separate threads in the Notification Center and on
        the lock screen. Grouped notifications are available beginning
        with iOS 12.

    >>> ios(alert='Hello!', sound='cat.caf',
    ...     extra={'articleid': '12345'}) # doctest: +SKIP
    {'sound': 'cat.caf', 'extra': {'articleid': '12345'}, 'alert': 'Hello!'}

    """
    payload = {}
    if alert is not None:
        if not (isinstance(alert, (string_type, dict))):
            raise ValueError('iOS alert must be a string or dictionary')
        payload['alert'] = alert
    if badge is not None:
        if not (isinstance(badge, (string_type, int))):
            raise ValueError('iOS badge must be an integer or string')
        if isinstance(badge, string_type) and not VALID_AUTOBADGE.match(badge):
            raise ValueError('Invalid iOS autobadge value')
        payload['badge'] = badge
    if sound is not None:
        if not (isinstance(sound, (string_type, dict))):
            raise ValueError('iOS sound must a string or dictionary')
        if isinstance(sound, dict) and 'name' not in sound:
            raise ValueError('if iOS sound is a dict, name key must be used')
        payload['sound'] = sound
    if content_available:
        payload['content-available'] = 1
    if extra is not None:
        payload['extra'] = extra
    if expiry is not None:
        if not (isinstance(expiry, (string_type, int))):
            raise ValueError('iOS expiry must be an integer or string')
        payload['expiry'] = expiry
    if interactive is not None:
        payload['interactive'] = interactive
    if category is not None:
        if not (isinstance(category, string_type)):
            raise ValueError('iOS category must be a string')
        payload['category'] = category
    if title is not None:
        if not (isinstance(title, string_type)):
            raise ValueError('iOS title must be a string')
        payload['title'] = title
    if mutable_content is not None:
        payload['mutable_content'] = mutable_content
    if subtitle is not None:
        payload['subtitle'] = subtitle
    if media_attachment is not None:
        payload['media_attachment'] = media_attachment
    if priority is not None:
        if priority not in {10, 5}:
            raise ValueError('iOS priority must be set to one of 5 or 10.')
        payload['priority'] = priority
    if collapse_id is not None:
        if not (isinstance(collapse_id, string_type)):
            raise ValueError('iOS collapse_id must be a string')
        payload['collapse_id'] = collapse_id
    if thread_id is not None:
        if not (isinstance(thread_id, string_type)):
            raise ValueError('iOS therad_id must be a string')
        payload['thread_id'] = thread_id

    return payload


def android(alert=None, collapse_key=None, time_to_live=None,
            delay_while_idle=False, extra=None, interactive=None,
            local_only=None, wearable=None, delivery_priority=None,
            style=None, title=None, summary=None, sound=None, priority=None,
            category=None, visibility=None, public_notification=None,
            notification_tag=None, notification_channel=None, icon=None,
            icon_color=None):
    """Android specific platform override payload.

    All keyword arguments are optional.

    :keyword alert: String alert text.
    :keyword collapse_key: String
    :keyword time_to_live: Integer
    :keyword delay_while_idle: Boolean
    :keyword extra: A set of key/value pairs to include in the push payload
        sent to the device. All values must be strings.
    :keyword interactive: A dictionary with two attributes -- "type" and
        "button_actions", as generated by :py:func:`interactive`.
    :keyword local_only: Optional value for not showing the notification on
        wearable devices. Defaults to False.
    :keyword wearable: Optional object to define a wearable notification
        with the following optional fields: background_image, extra_pages, and
        interactive.
    :keyword delivery_priority: Optional string of either 'high' or 'normal'.
        Sets the GCM priority.
    :keyword style: Optional object. Defines an advanced style.
    :keyword title: Optional string. Represents the title of the notification.
    :keyword summary: Optional string. Represents a summary of the
        notification.
    :keyword sound: Optional string. Represents a sound file name included
        in the app resources.
    :keyword priority: Optional integer between -2 and 2. An Android L feature
        that determines location sort order.
    :keyword category: Optional string. An Android category.
    :keyword visibility: Optional integer between -1 and 1.
    :keyword public_notification: Optional object. A notification to show on
        the lock screen instead instead of the redacted one.
    :keyword notification_tag: Optional string. A string identifier used to
        replace an existing notification in the notification drawer.
    :keyword notification_channel: Optional string.  An identifier of a
        notification channel predefined by the application.
    :keyword icon: Optional string. The name of a drawable in the
        application’s resource directory.
    :keyword icon_color: Optional. A string in the format of #rrggbb that
        defines the notification accent color.

    See
    `GCM Advanced Topics <http://developer.android.com/google/gcm/adv.html>`_
    for details on ``collapse_key``, ``time_to_live``, and
    ``delay_while_idle``.

    >>> android(alert='Hello!', extra={'articleid': '12345'}) # doctest: +SKIP
    {'extra': {'articleid': '12345'}, 'alert': 'Hello!'}

    """
    payload = {}
    if alert is not None:
        payload['alert'] = alert
    if collapse_key is not None:
        payload['collapse_key'] = collapse_key
    if time_to_live is not None:
        payload['time_to_live'] = time_to_live
        if not (isinstance(time_to_live, (string_type, int))):
            raise ValueError('Android time_to_live value must be an '
                             'integer or time set in UTC as a string')
    if delay_while_idle:
        payload['delay_while_idle'] = True
    if extra is not None:
        payload['extra'] = extra
    if interactive is not None:
        payload['interactive'] = interactive
    if local_only is not None:
        if not (isinstance(local_only, bool)):
            raise ValueError('Android local_only must be a boolean value')
        payload['local_only'] = local_only
    if wearable is not None:
        if not (isinstance(wearable, dict)):
            raise ValueError('Android wearable must be a dictionary')
        payload['wearable'] = wearable
    if delivery_priority is not None:
        if delivery_priority not in {'high', 'normal'}:
            raise ValueError(
                "delivery_priority must be set to one of 'high' or 'normal'."
            )
        payload["delivery_priority"] = delivery_priority
    if style is not None:
        payload['style'] = style
    if title is not None:
        payload['title'] = title
    if summary is not None:
        payload['summary'] = summary
    if sound is not None:
        payload['sound'] = sound
    if priority is not None:
        if priority not in range(-2, 3):
            raise ValueError(
                'priority must be set to one of {}.'.format(
                    ', '.join([str(i) for i in range(-2, 3)])
                )
            )
        payload['priority'] = priority
    if category is not None:
        if category not in VALID_ANDROID_CATEGORIES:
            raise ValueError(
                'category must be set to one of {}.'.format(
                    ', '.join(VALID_ANDROID_CATEGORIES)
                )
            )
        payload['category'] = category
    if visibility is not None:
        if visibility not in range(-1, 2):
            raise ValueError(
                'visibility must be set to one of {}.'.format(
                    ', '.join([str(i) for i in range(-1, 2)])
                )
            )
        payload['visibility'] = visibility
    if public_notification is not None:
        payload['public_notification'] = public_notification
    if notification_tag is not None:
        payload['notification_tag'] = notification_tag
    if notification_channel is not None:
        payload['notification_channel'] = notification_channel
    if icon is not None:
        payload['icon'] = icon
    if icon_color is not None:
        if (isinstance(icon_color, string_type) and not
                VALID_ICON_COLOR.match(icon_color)):
            raise ValueError('icon_color must be in format #rrggbb')
        payload['icon_color'] = icon_color

    return payload


def amazon(alert=None, consolidation_key=None, expires_after=None, extra=None,
           title=None, summary=None, interactive=None, style=None, sound=None):
    """Amazon specific platform override payload.

    All keyword arguments are optional.

    :keyword alert: String alert text.
    :keyword consolidation_key: String
    :keyword expires_after: Integer or UTC time (string)
    :keyword extra: A set of key/value pairs to include in the push payload
        sent to the device. All values must be strings.
    :keyword title: String
    :keyword summary: String
    :keyword interactive: A dictionary with two attributes -- "type" and
        "button_actions", as generated by :py:func:`interactive`.


    >>> amazon(alert='Hello!', title='My Title',
    ...     extra={'articleid': '12345'}) # doctest: +SKIP
    {'title': 'My Title', 'extra': {'articleid': '12345'}, 'alert': 'Hello!'}

    """
    payload = {}
    if alert is not None:
        payload['alert'] = alert
    if consolidation_key is not None:
        payload['consolidation_key'] = consolidation_key
    if expires_after is not None:
        payload['expires_after'] = expires_after
        if not (isinstance(expires_after, (string_type, int))):
            raise ValueError('Amazon time_to_live value must be an '
                             'integer or time set in UTC as a string')
    if extra is not None:
        payload['extra'] = extra
    if title is not None:
        payload['title'] = title
    if summary is not None:
        payload['summary'] = summary
    if interactive is not None:
        payload['interactive'] = interactive
    if style is not None:
        payload['style'] = style
    if sound is not None:
        payload['sound'] = sound
    return payload


def web(alert=None, extra=None, icon=None, title=None, interactive=None,
        time_to_live=None, require_interaction=None):
    """Web specific platform override payload.

    All keyword arguments are optional

    :keyword alert: An optional string alert text.
    :keyword extra: An optional set of key/value pairs to include in the push
        payload sent to the device.
    :keyword icon: An optional JSON dictionary of string key and value pairs.
        Key must be 'url' and value, an HTTPS URL to the icon resource.
    :keyword title: Optional string. Represents the title of the notification.
    :keyword interactive: An optional dictionary with two attributes --
        ``type`` and ``button_actions``, as generated by :py:func:`interactive`
    :keyword time_to_live: An optional integer or time set in UTC as a string
    :keyword require_interaction: An optional boolean value that requires user
        interaction on the notification before it is dismissed.


    >>> web(extra={'articleid': '12345'},
    ...    icon={'url': 'https://example.com/icon.png'},
    ...    title='This is a title!') # doctest: +SKIP
    {'title': 'This is a title!', 'extra': {'articleid': '12345'},
     'icon': {'url': 'https://example.com/icon.png'}, 'alert': 'Hello!'}

    """
    payload = {}
    if alert is not None:
        payload['alert'] = alert
    if extra is not None:
        payload['extra'] = extra
    if icon is not None:                # todo: possibly set 'url' key
        payload['icon'] = icon
        if not isinstance(icon, dict):
            raise TypeError('icon must be a dictionary')
    if title is not None:
        payload['title'] = title
    if interactive is not None:
        payload['interactive'] = interactive
    if time_to_live is not None:
        if not (isinstance(time_to_live, (string_type, int))):
            raise ValueError('Web time_to_live must be an integer or string')
        payload['time_to_live'] = time_to_live
    if require_interaction is not None:
        if not (isinstance(require_interaction, (bool))):
            raise ValueError('require_interaction must be boolean type')
        payload['require_interaction'] = require_interaction
    return payload


def sms(alert=None, expiry=None, template_alert=None):
    """Sms platform specific override payload

    One of alert or template_alert must be used.

    :keyword alert: An optional string. Overrides the alert provided at the
        top level of the notification for Sms channels.
        The maximum length of an Sms alert is 1600 characters.
    :keyword template_alert: for use with CreateAndSendPush. Substitutions must be added
        to the Sms objects passed to CreateAndSendPush that match the bracketed
        fields in this alert string.
    :keyword expiry: Optional. Delivery expiration, as either absolute
        UTC timestamp (string), or number of seconds from now (integer).


    >>> sms(alert='sms override alert!', expiry='2018-04-01T12:00:00') # doctest: +SKIP
    {'alert': 'sms override alert!', 'expiry': '2018-04-01T12:00:00'}
    """
    payload = {}
    if not alert and not template_alert:
        raise ValueError('One of alert and template_alert must be set.')
    elif alert and template_alert:
        raise ValueError('alert and template_alert cannot both be used.')

    if alert is not None:
        payload = {'alert': alert}
    elif template_alert is not None:
        payload = {'template': {'fields': {'alert': template_alert}}}
    
    if expiry is not None:
        if not (isinstance(expiry, (string_type, int))):
            raise ValueError('expiry value must be a string or integer')
        payload['expiry'] = expiry
    return payload


def email(message_type, plaintext_body, reply_to, sender_address,
          sender_name, subject, html_body=None, variable_defaults=None):
    """
    Required platform override when email in ua.device_types.
    Specifies content of the email being sent. All other notification
    fields are not used in the email if included.

    See https://docs.urbanairship.com/api/ua/#schemas%2femailoverrideobject
        for additional caveats.

    :param message_type: Required. One of transactional or commercial.
    :param plaintext_body: Required. The plain text body of the email. May include
        variable substitution fields for use with create and send inline templates.
    :param reply_to: Required. The reply-to address.
    :param sender_address: Required. The email address of the sender.
        The domain of the email must be enabled in the email
        provider at the delivery tier.
    :param sender_name: Required. The common name of the email sender.
    :param subject: Required. The subject line of the email. May include variable
        substitution fields for use with create and send inline templates.
    :param html_body: Optional. The HTML content of the email.
    :param variable_details: Required for CreateAndSendPush with inline templates.
        a list of dicts of default values for inline template variables.
    """
    if message_type not in ('transactional', 'commercial'):
        raise ValueError('message_type must be either transactional or commercial')

    payload = {}
    payload['message_type'] = message_type
    payload['reply_to'] = reply_to
    payload['sender_address'] = sender_address
    payload['sender_name'] = sender_name

    alert_payload = {
        'subject': subject,
        'plaintext_body': plaintext_body
    }
    if html_body is not None:
        alert_payload['html_body'] = html_body

    if variable_defaults is not None:
        payload['template'] = {
            'variable_defaults': variable_defaults,
            'fields': alert_payload
        }
    else:
        payload.update(alert_payload)

    return payload


def open_platform(alert=None, title=None, extra=None, summary=None,
                  media_attachment=None, interactive=None, template_alert=None):
    """Open platform specific override payload.

    All keyword arguments are optional.

    :keyword alert: String alert text.
    :keyword title: String, represents the title of the notification.
    :keyword summary: A string value for providing a content summary.
    :keyword media_attachment: A URI for an image or video somewhere on the
        internet.
    :keyword extra: A set of key/value pairs to include in the push payload
        sent to the device.
    :keyword interactive: An interactive payload override, as generated by
        :py:func:`interactive`. Included button actions must be of type
        ``add_tag``, ``remove tag``, ``app_defined``, or ``open`` with subtype
        ``url``.
    :keyword template_alert: For use with CreateAndSendPush. A string with
        inline template substitution fields. The name of these must match
        subsititutions on the OpenChannel objects used with CreateAndSendPush.

    >>> sms_overrides = open_platform(
    ...    alert='Hello!',
    ...    title='See my new homepage!',
    ...    summary='A longer summary of some content',
    ...    media_attachment='https://example.com/cat_standing_up.jpeg',
    ...    extra={'some_info': 'for sms only'},
    ...    interactive=interactive(
    ...            type='ua_yes_no_foreground',
    ...            button_actions={
    ...                'yes': actions(open_={
    ...                    'type':'url',
    ...                    'content':'https://www.urbanairship.com'
    ...                    }),
    ...                'no': actions(app_defined={'foo': 'bar'})
    ...            }
    ...        ) # doctest: +SKIP
    {'alert': 'Hello!', 'title': 'See my new homepage!', 'summary':
    'A longer summary of some content', 'media_attachment':
    'https://example.com/cat_standing_up.jpeg', 'extra': {'some_info':
    'for sms only'}, 'interactive': {'type': 'ua_yes_no_foreground',
    'button_actions': { 'yes': {'open': {'type': 'url', 'content':
    'http://www.urbanairship.com'}}, 'no': {'app_defined': {'foo': 'bar'}}}}}

    >>> notification(open_platform={'sms': sms_overrides}) # doctest: +SKIP

    """
    payload = {}
    if extra is not None:
        payload['extra'] = extra
    if interactive is not None:
        payload['interactive'] = interactive

    alert_payload = {}
    if title is not None:
        alert_payload['title'] = title
    if summary is not None:
        alert_payload['summary'] = summary
    if media_attachment is not None:
        alert_payload['media_attachment'] = media_attachment

    if template_alert is not None:
        alert_payload['alert'] = template_alert
        payload = {'template': {'fields': alert_payload}}
    else:
        alert_payload['alert'] = alert
        payload.update(alert_payload)

    return payload


def wns_payload(alert=None, toast=None, tile=None, badge=None):
    """WNS specific platform override payload.

    Must include exactly one of ``alert``, ``toast``, ``tile``, or ``badge``.

    """
    if sum(1 for x in (alert, toast, tile, badge) if x) != 1:
        raise ValueError('WNS payload must have one notification type.')
    payload = {}
    if alert is not None:
        payload['alert'] = alert
    if toast is not None:
        payload['toast'] = toast
    if tile is not None:
        payload['tile'] = tile
    if badge is not None:
        payload['badge'] = badge
    return payload


def message(title, body, content_type=None, content_encoding=None,
            extra=None, expiry=None, icons=None, options=None,
            campaigns=None):
    """Rich push message payload creation.

    :param title: Required, string
    :param body: Required, string
    :keyword content_type: Optional, MIME type of the body
    :keyword content_encoding: Optional, encoding of the data in body,
        e.g. ``utf-8``.
    :keyword extra: Optional, dictionary of string values.
    :keyword expiry: time when message will delete from Inbox
        (UTC time or in seconds)
    :keyword icons: Optional JSON dictionary of string key and value
        pairs. Values must be URIs or URLs to the icon resources
    :keyword options: Optional JSON dictionary of key and value pairs
        specifying non-payload options
    :keyword campaigns: Optional dictionary containing list of 1 to
        10 'categories'

    """
    payload = {
        'title': title,
        'body': body,
    }
    if content_type is not None:
        payload['content_type'] = content_type
    if content_encoding is not None:
        payload['content_encoding'] = content_encoding
    if extra is not None:
        payload['extra'] = extra
    if expiry is not None:
        payload['expiry'] = expiry
        if not (isinstance(expiry, (string_type, int))):
            raise ValueError('Expiry value must be an '
                             'integer or time set in UTC as a string')
    if icons is not None:
        if not isinstance(icons, dict):
            raise TypeError('icons must be a dictionary')
        payload['icons'] = icons
    if options is not None:
        if not isinstance(options, dict):
            raise TypeError('options must be a dictionary')
        payload['options'] = options
    if campaigns is not None:
        if not isinstance(campaigns, dict):
            raise TypeError('campaigns must be a dictionary')
        payload['campaigns'] = campaigns

    return payload


def in_app(alert, display_type, expiry=None, display=None,
           actions=None, interactive=None, extra=None):
    """In-App push message payload creation.

    :param alert: Required, string
    :param display_type: Required, String. List of valid display_type_
        attributes.
    :keyword expiry: Optional, String specifying an expiry value.
    :keyword display: Optional, A dictionary specifying the display_
        appearance of the in-app message.
    :keyword actions: Optional, Used to perform a defined task.
    :keyword interactive: Optional, A dictionary with two attributes –
        "type" and "button_actions", as generated by interactive().
    :keyword extra: Optional, Object.

    .. _display: https://docs.urbanairship.com/api/ua.html#display-object
    .. _display_type: https://docs.urbanairship.com/api/ua.html#in-app-message

    """

    payload = {}

    payload['alert'] = alert
    payload['display_type'] = display_type

    if expiry is not None:
        payload['expiry'] = expiry
    if display is not None:
        payload['display'] = display
    if actions is not None:
        payload['actions'] = actions
    if interactive is not None:
        payload['interactive'] = interactive
    if extra is not None:
        extra['extra'] = extra

    return payload


def device_types(*types):
    """Create a device type specifier.

    >>> device_types('ios', 'wns')
    ['ios', 'wns']
    >>> device_types('ios', 'symbian')
    Traceback (most recent call last):
        ...
    ValueError: Invalid device type 'symbian'

    """
    valid_device_types = (
        'ios',
        'android',
        'amazon',
        'wns',
        'web',
        'sms',
        'email'
    )
    if types == 'all' or (len(types) == 1 and types[0] == 'all'):
        warnings.warn(
            "The device type 'all' has been deprecated.",
            DeprecationWarning
        )
        return 'all'

    for t in types:
        is_open = isinstance(t, string_type) and t.startswith('open::')
        if t not in valid_device_types and not is_open:
            raise ValueError("Invalid device type '%s'" % t)
    return [t for t in types]


def options(expiry=None):
    """Options payload creation.

    :keyword expiry: time at which push will no longer be sent.
        Int or UTC time

    """
    payload = {}
    if expiry is not None:
        payload['expiry'] = expiry
    if not (isinstance(expiry, (string_type, int))):
        raise ValueError('Expiry value must be an '
                         'integer or time set in UTC as a string')
    return payload


def campaigns(categories=None):
    """Campaigns payload creation.

    >>> campaigns(categories=['kittens', 'tacos', 'horse_racing'])
    {'categories': ['kittens', 'tacos', 'horse_racing']}

    """
    payload = {}
    if categories is not None:
        if not isinstance(categories, collections.Sequence):
            raise TypeError('Each category must be a string')
        if isinstance(categories, list):
            if not categories or len(categories) > 10:
                raise ValueError(
                    'Categories list must contain between 1 and 10 items'
                )
        if isinstance(categories, string_type):
            categories = [categories]

        for c in categories:
            if not (isinstance(c, string_type)):
                raise ValueError("Invalid category type '%s'" % c)
            if not len(c) or len(c) > 64:
                raise ValueError("Invalid category name '%s'" % c)
        payload['categories'] = [c for c in categories]
    return payload


def actions(add_tag=None, remove_tag=None,
            open_=None, share=None, app_defined=None):
    """Actions payload creation.

    :keyword add_tag: Adds a tag to the device. Expects a
        string or a list of strings.
    :keyword remove_tag: Removes a tag from the device. Expects
        a string or a list of strings.
    :keyword open_: Opens type url, deep_link or landing_page. Expects a
        dictionary with "type" and "content".
        See API docs for more information.
    :keyword share: Sends a share notification. Expects a string.
    :keyword app_defined: Sends application defined actions. Expects
        a dictionary.

    >>> actions(add_tag='new_tag', remove_tag='old_tag',
    ...     open_={'type': 'url',
    ...         'content': 'http://www.urbanairship.com'}) # doctest: +SKIP
    {'open': {'type': 'url', 'content': 'http://www.urbanairship.com},
     'add_tag': 'new_tag', 'remove_tag': 'old_tag'}

    """
    payload = {}
    if add_tag is not None:
        if not (isinstance(add_tag, (collections.Sequence))):
            raise TypeError('add_tag must be a string or a list of strings')
        if isinstance(add_tag, list) and not add_tag:
            raise ValueError('add_tag list cannot be empty')
        payload['add_tag'] = add_tag
    if remove_tag is not None:
        if not (isinstance(remove_tag, (collections.Sequence))):
            raise TypeError('remove_tag must be a string or a list of strings')
        if isinstance(remove_tag, list) and not remove_tag:
            raise ValueError('remove_tag list cannot be empty')
        payload['remove_tag'] = remove_tag
    if open_ is not None:
        if not (isinstance(open_, dict)):
            raise TypeError('open_ must be a dictionary')
        payload['open'] = open_
    if share is not None:
        if not (isinstance(share, string_type)):
            raise TypeError('share must be a string')
        payload['share'] = share
    if app_defined is not None:
        if not (isinstance(app_defined, dict)):
            raise TypeError('app_defined must be a dictionary')
        payload['app_defined'] = app_defined
    return payload


def interactive(type=None, button_actions=None):
    """Interactive payload creation.
    :keyword type: The name of one of the predefined interactive notifications
    or a custom defined interactive notification. Expects a string.
    :keyword button_actions: A button_actions object that maps button IDs to
    valid action objects. Expects a dictionary.
    """

    payload = {}
    if type is not None:
        payload['type'] = type
        if button_actions is not None:
            if not isinstance(button_actions, dict):
                raise TypeError("'button_actions' must be a dictionary")
            payload['button_actions'] = button_actions
    else:
        raise AttributeError("'interactive' must have a type attribute")

    return payload


def wearable(background_image=None, extra_pages=None, interactive=None):
    """Android wearable payload builder.

    :keyword background_image: Optional string. A URL that specifies the
        background image to display on a wearable device.
    :keyword extra_pages: Optional array of objects.
    :keyword interactive: Optional object. Override the interactive
        notification payload for the wearable device.
    """
    payload = {
        'background_image': background_image,
        'extra_pages': extra_pages,
        'interactive': interactive
    }
    return {key: val for key, val in iter(payload.items()) if val is not None}


def public_notification(title=None, alert=None, summary=None):
    """Android L public notification payload builder.

    :keyword title: Optional string. The notification title.
    :keyword alert: Optional string. The notification alert.
    :keyword summary: Optional string. The notification summary.
    """
    payload = {'title': title, 'alert': alert, 'summary': summary}
    return {key: val for key, val in iter(payload.items()) if val is not None}


def style(style_type, content, title=None, summary=None):
    """Android/Amazon style builder.

    :keyword style_type: String. Must be one of "big_text", "big_picture",
        or "inbox".
    :keyword content: String or array of strings. Content of the style object.
        If style_type is set to "inbox", this will be an array of strings.
        Otherwise, it will be a single string.
    :keyword title: Optional string. Override the notification.
    :keyword summary: Optional string. Override the summary of the
        notification.

    """
    mapping = {
        'big_text': 'big_text', 'big_picture': 'big_picture', 'inbox': 'lines'
    }
    if style_type not in mapping.keys():
        raise ValueError('style_type must be one of {}.'.format(
            ', '.join(mapping.keys())
        ))
    payload = {
        'type': style_type, mapping[style_type]: content,
        'title': title, 'summary': summary
    }
    return {key: val for key, val in iter(payload.items()) if val is not None}


def media_attachment(url, content=None, options=None):
    """iOS media_attachment builder.

    :keyword url: String. Specifies the URL to be downloaded by the UA
        Media Attachment extension.
    :keyword content: Optional dictionary. Describes portions of the
        notification that should be modified if the media attachment
        succeeds. See :func:`content`.
    :keyword options: Optional dictionary. Describes how to display the
        resource given by the URL. See :func:`options`.
    """
    payload = {'url': url, 'content': content, 'options': options}
    return {key: val for key, val in iter(payload.items()) if val is not None}


def content(title=None, subtitle=None, body=None):
    """iOS content builder. Each argument describes the portions of the
    notifcation that should be modified if the media_attachment succeeds.

    :keyword title: String.
    :keyword subtitle: String.
    :keyword body: String.
    """
    payload = {'title': title, 'subtitle': subtitle, 'options': options}
    return {key: val for key, val in iter(payload.items()) if val is not None}


def crop(x=None, y=None, width=None, height=None):
    """iOS crop builder.

    :keyword x: Optional float. The X offset where the crop begins.
    :keyword y: Optional float. The Y offset where the crop begins.
    :keyword width: Optional float. The width of the final crop.
    :keyword height: Optional float. The height of the final crop.
    """
    payload = {'x': x, 'y': y, 'width': width, 'height': height}
    return {key: val for key, val in iter(payload.items()) if val is not None}



class Pipeline(object):
    """Pipeline object encapsulates the complete set of objects that define
    an Automation pipeline
    """

    def __init__(self,
                 enabled=None,
                 outcome=[],
                 name=None,
                 immediate_trigger=[],
                 cancellation_trigger=[],
                 historical_trigger=None,
                 constraint=[],
                 condition=[],
                 timing=None):
        """Create Pipeline object for Automation payload

        :keyword enabled: boolean value determines active status of the
        pipeline
        :keyword outcome: a push object to be sent as outcome of the pipeline
        :keyword name: [optional] descriptive string for pipeline
        identification
        :keyword immediate_trigger: [optional] single event identifier or list
            of event identifiers that trigger the outcome of the pipeline -
            trigger list can be combination of simple and compound identifiers
        :keyword cancellation_trigger: [optional] single event identifier or
            list of event identifiers that trigger cancellation of the pipeline
            outcome - can be combination of simple and compound identifiers
        :keyword historical_trigger: [optional] single or list of historical
            trigger objects that currently match only inactive triggers
        :keyword constraint: [optional] single or list of constraint objects
            that limit the number of pushes sent to a device in a given time
            period
        :keyword condition: [optional] one or a list of condition sets with an
            and/or operator for combining the conditions
        :keyword timing: [optional] a timing object defining times allowable
            for outcome delivery
        """
        self.enabled = enabled
        self.outcome = outcome
        self.name = name
        self.immediate_trigger = immediate_trigger
        self.cancellation_trigger = cancellation_trigger
        self.historical_trigger = historical_trigger
        self.constraint = constraint
        self.condition = condition
        self.timing = timing

    @property
    def payload(self):
        """JSON serialized Pipeline object"""
        if not isinstance(self.enabled, bool):
            raise TypeError('enabled is required and must be a boolean value')
        if not self.outcome:
            raise ValueError('outcome is required for automation pipelines')

        data = {
            "enabled": self.enabled,
            "outcome": self.outcome
        }

        if self.name is not None:
            if isinstance(self.name, str):
                data['name'] = self.name
        if self.immediate_trigger is not None:
            data['immediate_trigger'] = self.immediate_trigger
        if self.cancellation_trigger is not None:
            data['cancellation_trigger'] = self.cancellation_trigger
        if self.historical_trigger is not None:
            data['historical_trigger'] = self.historical_trigger
        if self.constraint is not None:
            data['constraint'] = self.constraint
        if self.condition is not None:
            data['condition'] = self.condition
        if self.timing is not None:
            data['timing'] = self.timing

        return data

    def from_dict(self, pipeline_dict):
        for k in pipeline_dict:
            setattr(self, k, pipeline_dict[k])

    @property
    def outcome(self):
        if not self._outcomes:
            return None
        if len(self._outcomes) == 1:
            return self._outcomes[0]

        return self._outcomes

    @outcome.setter
    def outcome(self, outcome_object):
        """Set outcome object

        :keyword outcome_object: Single outcome object or list of outcome objects;
        outcomes are push objects
        """
        if outcome_object is None:
            self._outcomes = []
        elif isinstance(outcome_object, list):
            to_set = []
            for push in outcome_object:
                if 'push' in push:
                    to_set.append(push)
                else:
                    to_set.append({'push': push})
            self._outcomes = to_set
        elif isinstance(outcome_object, dict):
            if 'push' in outcome_object:
                self._outcomes = [outcome_object]
            else:
                self._outcomes = [{'push': outcome_object}]
        else:
            raise TypeError(
                    'outcome must be outcome object or list of outcome objects'
            )

    def append_outcome_object(self, push_object):
        """Append outcome object to current Pipeline outcome

        :keyword push_object: A push object that will be an outcome for this
        Pipeline object
        """
        if not isinstance(push_object, dict):
            TypeError(
                'outcome object requires a push object as a dictionary'
            )

        to_append = {}
        if isinstance(push_object, dict):
            to_append = push_object

        self._outcomes.append(to_append)

    def remove_outcome_object(self, push_object):
        """Remove outcome push object from current outcome

        :keyword push_object: A push object to remove from the outcome of this
        Pipeline object
        """
        self._outcomes.remove(push_object)

    @property
    def immediate_trigger(self):
        if not self._immediate_trigger:
            return None
        if len(self._immediate_trigger) == 1:
            return self._immediate_trigger[0]

        return self._immediate_trigger

    @immediate_trigger.setter
    def immediate_trigger(self, event_identifiers):
        """Set immediate trigger for Pipeline

        :keyword event_identifiers: One or list of event identifiers
        """
        if event_identifiers:
            if not isinstance(event_identifiers, (dict, list)):
                TypeError(
                    'immediate trigger must be an event identifier dictionary '
                    'or list of event identifier dictionaries'
                )

            if isinstance(event_identifiers, list):
                self._immediate_trigger = event_identifiers

            if isinstance(event_identifiers, dict):
                self._immediate_trigger = [event_identifiers]
        else:
            self._immediate_trigger = []

    def append_immediate_trigger_identifier(self, event_identifier):
        """Append event identifier to immediate triggers for Pipeline

        :keyword event_identifier: Even identifier object
        """
        self._immediate_trigger.append(event_identifier)

    def remove_immediate_trigger_identifier(self, event_identifier):
        """Remove event identifier to immediate triggers for Pipeline

        :keyword event_identifier: Even identifier object
        """
        self._immediate_trigger.remove(event_identifier)

    @property
    def cancellation_trigger(self):
        if not self._cancellation_trigger:
            return None
        if len(self._cancellation_trigger) == 1:
            return self._cancellation_trigger[0]

        return self._cancellation_trigger

    @cancellation_trigger.setter
    def cancellation_trigger(self, event_identifiers):
        """Set Pipeline cancellation trigger

        :keyword event_identifiers: One of list of event identifiers
        """
        if event_identifiers:
            if not isinstance(event_identifiers, (str, dict, list)):
                TypeError(
                    'immediate trigger must be an event identifier or '
                    'list of event identifiers'
                )

            if isinstance(event_identifiers, list):
                self._cancellation_trigger = event_identifiers
            else:
                self._cancellation_trigger = [event_identifiers]
        else:
            self._cancellation_trigger = []

    def append_cancellation_trigger_identifier(self, event_identifier):
        """Append event identifier to immediate triggers for Pipeline

        :keyword event_identifier: Event identifier object
        """
        self._cancellation_trigger.append(event_identifier)

    def remove_cancellation_trigger_identifier(self, event_identifier):
        """Remove event identifier from immediate triggers for Pipeline

        :keyword event_identifier: Event identifier object
        """
        self._cancellation_trigger.remove(event_identifier)

    @property
    def historical_trigger(self):
        return self._historical_trigger

    @historical_trigger.setter
    def historical_trigger(self, historical_trigger_object):
        """Set historical trigger for Pipeline

        :keyword historical_trigger_object: Historical trigger object for
        triggering Pipeline outcome
        """
        if historical_trigger_object:
            if historical_trigger_object['event'] != "open":
                raise ValueError(
                    'only allowable value of historical trigger "event" is '
                    '"open"'
                )

            if historical_trigger_object['equals'] != 0:
                raise ValueError(
                    'only allowable value of historial trigger "equals" is 0'
                )

            if 0 >= historical_trigger_object['days'] > 90:
                raise ValueError('days of inactivity must be between 1 and 90')

        self._historical_trigger = historical_trigger_object

    @property
    def constraint(self):
        if not self._constraints:
            return None
        if len(self._constraints) == 1:
            return self._constraints[0]

        return self._constraints

    @constraint.setter
    def constraint(self, constraint_objects):
        """Set Pipline constraints

        :keyword constraint_objects: One of list of constraint objects
        """
        if constraint_objects:
            if not isinstance(constraint_objects, (dict, list)):
                TypeError(
                    'constraint must be single constraint object or list of '
                    'constraint objects'
                )

            if isinstance(constraint_objects, list):
                self._constraints = constraint_objects
            if isinstance(constraint_objects, dict):
                self._constraints = [constraint_objects]
        else:
            self._constraints = []

    def append_constraint_object(self, constraint_object):
        """Append constraint object to constraint for Pipeline

        :keyword constraint_object: Constraint object to append to Pipeline
        constraint
        """
        if not isinstance(constraint_object, dict):
            TypeError('a single constraint object must be a dictionary')

        self._constraints.append(constraint_object)

    def remove_constraint_object(self, constraint_object):
        self._constraints.remove(constraint_object)

    @property
    def condition(self):
        if not self._condition_sets:
            return None
        if len(self._condition_sets) == 1:
            return self._condition_sets[0]

        return self._condition_sets

    @condition.setter
    def condition(self, condition_sets):
        """Set condition for Pipeline

        :keyword condition_sets: One or list of condition sets
        """
        if condition_sets:
            if not isinstance(condition_sets, (list, dict)):
                raise TypeError(
                    'condition_sets must be a condition set '
                    'or list of condition sets'
                )

            if isinstance(condition_sets, list):
                if len(condition_sets) > 20:
                    raise ValueError('condition maximum is 20 condition sets')

            if self._validate_condition_set(condition_sets):
                if isinstance(condition_sets, list):
                    self._condition_sets = condition_sets
                else:
                    self._condition_sets = [condition_sets]
        else:
            self._condition_sets = []

    def append_condition_set(self, condition_set):
        """Append condition set to Pipeline condition

        :keyword condition_set: One or list of condition sets
        """
        if not isinstance(condition_set, dict):
            TypeError('a single condition set object must be a dictionary')

        if self._validate_condition_set(condition_set):
            self._condition_sets.append(condition_set)

    def remove_condition_set(self, condition_set):
        """Remove condition set from Pipeline condition

        :keyword condition_set: One or list of condition sets
        """
        self._condition_sets.remove(condition_set)

    def _validate_condition_set(self, condition_sets):
        invalid_operators = []
        if isinstance(condition_sets, list):
            for condition_set in condition_sets:
                for operator in condition_set:
                    if operator not in ['and', 'or']:
                        invalid_operators.append(operator)
        else:
            for operator in condition_sets:
                if operator not in ['and', 'or']:
                    invalid_operators.append(operator)

        if invalid_operators:
            raise KeyError(
                'invalid operators: {}, must be "and" or "or"'.format(
                    ', '.join(invalid_operators)
                )
            )

        return True

    @property
    def timing(self):
        return self._timing

    @timing.setter
    def timing(self, timing_object):
        """Set timing object for available send times for Pipeline

        :keyword timing_object: Single timing object defining allowable
        delivery times for Pipeline
        """
        if timing_object:
            if 'delay' in timing_object:
                delay = timing_object['delay']['seconds']

                if not (isinstance(delay, int) and delay > 0):
                    raise ValueError('delay must be an integer greater than 0')

            if 'schedule' in timing_object:
                schedule = timing_object['schedule']
                if 'type' not in schedule:
                    raise KeyError('"type" is required for a timing schedule')
                if not schedule['type'] in ('local', 'utc'):
                    raise ValueError(
                        'timing schedule "type" must be "local" or "utc"'
                    )

                if 'dayparts' not in schedule:
                    raise KeyError(
                        '"dayparts" is required for a timing schedule'
                    )

            self._timing = timing_object
        else:
            self._timing = None



import json
import logging

from urbanairship import common

logger = logging.getLogger('urbanairship')


class Segment(object):
    _airship = None
    url = None

    id = None
    display_name = None
    creation_date = None
    modification_date = None
    criteria = None
    data = None

    def create(self, airship):
        """Create a Segment object and return it."""

        url = airship.urls.get('segments_url')

        body = json.dumps(
            {
                'display_name': self.display_name,
                'criteria': self.criteria
            }
        )
        response = airship._request(
            method='POST',
            body=body,
            url=url,
            version=3
        )
        logger.info(
            'Successful segment creation: {0}'.format(self.display_name)
        )

        seg_url = response.headers['location']
        seg_id = seg_url.split(url)[1]

        self.id = seg_id
        self.from_id(airship, seg_id)

        return response

    @classmethod
    def from_id(cls, airship, seg_id):
        """Retrieve a segment based on the provided ID."""

        url = airship.urls.get('segments_url') + seg_id
        response = airship._request(
            method='GET',
            body=None,
            url=url,
            version=3
        )

        payload = response.json()
        cls.id = seg_id
        cls.from_payload(payload)

        return response

    @classmethod
    def from_payload(cls, payload):
        """Create segment based on results from a SegmentList iterator."""

        for key in payload:
            setattr(cls, key, payload[key])

        return cls

    def update(self, airship):
        """Updates the segment associated with data in the current object."""

        data = {}
        data['display_name'] = self.display_name
        data['criteria'] = self.criteria

        url = airship.urls.get('segments_url') + self.id
        body = json.dumps(data)
        response = airship._request(
            method='PUT',
            body=body,
            url=url,
            version=3
        )
        logger.info(
            "Successful segment update: '{0}'".format(self.display_name))

        return response

    def delete(self, airship):
        url = airship.urls.get('segments_url') + self.id
        res = airship._request(
            method='DELETE',
            body=None,
            url=url,
            version=3
        )
        logger.info(
            "Successful segment deletion: '{0}'".format(self.display_name))
        return res


class SegmentList(common.IteratorParent):
    """Retrieves a list of segments

        :ivar limit: Number of segments to fetch

    """
    next_url = None
    data_attribute = 'segments'

    def __init__(self, airship, limit=None):
        self.next_url = airship.urls.get('segments_url')
        params = {'limit': limit} if limit else {}
        super(SegmentList, self).__init__(airship, params)



"""Python package for using the Urban Airship API"""
from .core import Airship
from .common import AirshipFailure, Unauthorized
from .push import (
    Push,
    ScheduledPush,
    ScheduledList,
    TemplatePush,
    Template,
    TemplateList,
    CreateAndSendPush,
    all_,
    ios_channel,
    android_channel,
    amazon_channel,
    channel,
    open_channel,
    sms_id,
    sms_sender,
    device_token,
    apid,
    wns,
    tag,
    alias,
    segment,
    and_,
    or_,
    not_,
    location,
    recent_date,
    absolute_date,
    notification,
    ios,
    android,
    amazon,
    web,
    sms,
    email,
    wns_payload,
    open_platform,
    message,
    in_app,
    device_types,
    options,
    campaigns,
    actions,
    interactive,
    wearable,
    public_notification,
    style,
    scheduled_time,
    local_scheduled_time,
    best_time,
    named_user,
    merge_data,
)

from .automation import (
    Automation,
    Pipeline
)

from .devices import (
    ChannelList,
    ChannelInfo,
    OpenChannel,
    DeviceTokenList,
    APIDList,
    DeviceInfo,
    ChannelTags,
    OpenChannelTags,
    Segment,
    SegmentList,
    ChannelUninstall,
    NamedUser,
    NamedUserList,
    NamedUserTags,
    StaticList,
    StaticLists,
    LocationFinder,
    Sms,
    Email,
    EmailTags,
)

from .reports import (
    IndividualResponseStats,
    ResponseList,
    DevicesReport,
    OptInList,
    OptOutList,
    PushList,
    ResponseReportList,
    AppOpensList,
    TimeInAppList,
    CustomEventsList,
)

__all__ = [
    Airship,
    AirshipFailure,
    Unauthorized,
    all_,
    Push,
    ScheduledPush,
    TemplatePush,
    ios_channel,
    android_channel,
    amazon_channel,
    channel,
    open_channel,
    device_token,
    apid,
    wns,
    tag,
    alias,
    segment,
    sms_id,
    sms_sender,
    and_,
    or_,
    not_,
    location,
    recent_date,
    absolute_date,
    notification,
    ios,
    android,
    amazon,
    web,
    wns_payload,
    open_platform,
    message,
    in_app,
    options,
    campaigns,
    actions,
    interactive,
    device_types,
    scheduled_time,
    local_scheduled_time,
    ChannelList,
    ChannelInfo,
    OpenChannel,
    Sms,
    DeviceTokenList,
    APIDList,
    DeviceInfo,
    Segment,
    SegmentList,
    ChannelUninstall,
    NamedUser,
    NamedUserList,
    NamedUserTags,
    IndividualResponseStats,
    ResponseList,
    DevicesReport,
    OptInList,
    OptOutList,
    PushList,
    ResponseReportList,
    AppOpensList,
    TimeInAppList,
    CustomEventsList,
    StaticList,
    StaticLists,
    LocationFinder,
    named_user,
    merge_data,
    Template,
    TemplateList,
    ScheduledList,
    Automation,
    Pipeline,
    Email,
    EmailTags,
    CreateAndSendPush
]

# Silence urllib3 INFO logging by default
from .experiments import (
    Experiment,
    Variant,
    ABTest
)

import logging
logging.getLogger('requests.packages.urllib3.connectionpool')\
    .setLevel(logging.WARNING)



class Variant(object):
    """The variants for the experiment. An experiment must have at least 1 variant
    and no more than 26.
    """

    def __init__(self,
                 push,
                 description=None,
                 name=None,
                 schedule=None,
                 weight=None
                 ):
        """
        :keyword push: [required] A push object without audience and device_types
            fields. These two fields are not allowed because they are already defined
            in the experiment object
        :keyword description: [optional] A description of the variant.
        :keyword name: [optional] A name for the variant
            unless either message or in_app is present. You can provide an alert and any
            platform overrides that apply to the device_type platforms you specify.
        :keyword schedule: [optional] The time when the push notification should be sent
        :keyword weight: [optional] The proportion of the audience that will receive
            this variant. Defaults to 1.
        """
        self.push = push
        self.description = description
        self.name = name
        self.schedule = schedule
        self.weight = weight

    @property
    def description(self):
        if not self._description:
            return None
        return self._description

    @description.setter
    def description(self, value):
        if not isinstance(value, str):
            TypeError(
                'the description must be type string'
            )

        self._description = value

    @property
    def name(self):
        if not self._name:
            return None
        return self._name

    @name.setter
    def name(self, value):
        if not isinstance(value, str):
            TypeError(
                'the name must be a string type'
            )

        self._name = value

    @property
    def weight(self):
        if not self._weight:
            return None
        return self._weight

    @weight.setter
    def weight(self, value):
        if not isinstance(value, int):
            TypeError(
                'the value must be a integer type'
            )
        self._weight = value



from .devicelist import (
    ChannelList,
    ChannelInfo,
    DeviceTokenList,
    APIDList,
    DeviceInfo,
)

from .tag import (
    ChannelTags,
    OpenChannelTags
)

from .segment import (
    Segment,
    SegmentList
)

from .channel_uninstall import (
    ChannelUninstall
)

from .open_channel import (
    OpenChannel
)

from .named_users import (
    NamedUser,
    NamedUserList,
    NamedUserTags
)

from .static_lists import (
    StaticList,
    StaticLists,
)

from .locationfinder import (
    LocationFinder
)

from .sms import (
    Sms
)

from .email import (
    Email,
    EmailTags
)



import json
import gzip
import collections
import datetime

from urbanairship import common

CHUNK = 16 * 1024


class StaticList(object):
    def __init__(self, airship, name):
        self.airship = airship
        self.name = name
        self.description = None
        self.extra = None

    def create(self):
        payload = {'name': self.name}
        if self.description is not None:
            payload['description'] = self.description
        if self.extra is not None:
            payload['extra'] = self.extra

        body = json.dumps(payload).encode('utf-8')
        response = self.airship._request(
            'POST',
            body,
            self.airship.urls.get('lists_url'),
            'application/json',
            version=3
        )
        return response.json()

    def upload(self, csv_file):
        """Upload a CSV file to a static list

        :param csv_file: open file descriptor with two column format:
            identifier_type, identifier
        :return: http response
        """

        zipped = GzipCompressReadStream(csv_file)
        url = self.airship.urls.get('lists_url') + self.name + '/csv/'
        response = self.airship._request(
            method='PUT',
            body=zipped,
            url=url,
            content_type='text/csv',
            version=3,
            encoding='gzip'
        )
        return response.json()

    def update(self):
        """Update the metadata in a static list
        :return: http response
        """

        if self.description is None and self.extra is None:
            raise ValueError('Either description or extra must be non-empty.')

        payload = {}

        if self.description is not None:
            payload['description'] = self.description
        if self.extra is not None:
            payload['extra'] = self.extra

        body = json.dumps(payload).encode('utf-8')
        url = self.airship.urls.get('lists_url') + self.name

        response = self.airship._request(
            'PUT', body, url, 'application/json', version=3
        )

        return response.json()

    @classmethod
    def from_payload(cls, payload, airship):
        obj = cls(airship, payload['name'])
        for key in payload:
            if key in 'created' or key in 'last_updated':
                payload[key] = datetime.datetime.strptime(
                    payload[key], '%Y-%m-%dT%H:%M:%S'
                )
            setattr(obj, key, payload[key])
        return obj

    def lookup(self):
        """
        :return: Information about the static list
        """

        url = self.airship.urls.get('lists_url') + self.name
        response = self.airship._request('GET', None, url, version=3)
        payload = response.json()
        return self.from_payload(payload, self.airship)

    def delete(self):
        """
        :return: Delete the static list
        """
        url = self.airship.urls.get('lists_url') + self.name
        return self.airship._request('DELETE', None, url, version=3)


class StaticLists(common.IteratorParent):
    next_url = None
    data_attribute = 'lists'

    def __init__(self, airship):
        self.next_url = airship.urls.get('lists_url')
        super(StaticLists, self).__init__(airship, None)


class Buffer(object):
    def __init__(self):
        self.__buf = collections.deque()
        self.__size = 0

    def __len__(self):
        return self.__size

    def write(self, data):
        self.__buf.append(data)
        self.__size += len(data)

    def read(self, size):
        ret_list = []
        while size > 0 and len(self.__buf):
            chunk_of_file = self.__buf.popleft()
            size -= len(chunk_of_file)
            ret_list.append(chunk_of_file)
        if size < 0:
            ret_list[-1], remainder = ret_list[-1][:size], ret_list[-1][size:]
            self.__buf.appendleft(remainder)
        ret = b''.join(ret_list)
        self.__size -= len(ret)
        return ret

    def flush(self):
        pass

    def close(self):
        pass


class GzipCompressReadStream (object):
    def __init__(self, file_obj):
        self.__input = file_obj
        self.__buf = Buffer()
        self.__gzip = gzip.GzipFile(None, mode='wb', fileobj=self.__buf)
        self.is_finished = False

    def read(self, size):
        while len(self.__buf) < size:
            chunk_of_file = self.__input.read(CHUNK)
            if not chunk_of_file:
                self.__gzip.close()
                self.is_finished = True
                break
            self.__gzip.write(chunk_of_file)
            self.__gzip.flush()
        return self.__buf.read(size)

    def __iter__(self):
        return self

    def __next__(self):
        if self.is_finished is True:
            raise StopIteration
        else:
            return self.read(CHUNK)

    def next(self):
        return self.__next__()



from .core import (
    Push, 
    ScheduledPush, 
    TemplatePush, 
    CreateAndSendPush,
)

from .audience import (
    ios_channel,
    android_channel,
    amazon_channel,
    channel,
    open_channel,
    sms_id,
    sms_sender,
    device_token,
    apid,
    wns,
    tag,
    alias,
    segment,
    and_,
    or_,
    not_,
    location,
    recent_date,
    absolute_date,
    named_user,
)

from .payload import (
    notification,
    ios,
    android,
    amazon,
    wns_payload,
    web,
    sms,
    email,
    open_platform,
    message,
    device_types,
    options,
    campaigns,
    actions,
    interactive,
    in_app,
    wearable,
    style,
    public_notification,
)

from .schedule import (
    scheduled_time,
    local_scheduled_time,
    best_time,
    ScheduledList,
)

from .template import (
    merge_data,
    Template,
    TemplateList,
)

# Common selector for audience & device_types

all_ = "all"
"""Select all, to do a broadcast.

Used in both ``audience`` and ``device_types``.
"""


__all__ = [
    all_,
    Push,
    ScheduledPush,
    ScheduledList,
    TemplatePush,
    Template,
    TemplateList,
    ios_channel,
    android_channel,
    amazon_channel,
    channel,
    open_channel,
    sms_id,
    sms_sender,
    device_token,
    apid,
    wns,
    tag,
    alias,
    segment,
    and_,
    or_,
    not_,
    recent_date,
    absolute_date,
    notification,
    ios,
    android,
    amazon,
    web,
    sms,
    wns_payload,
    open_platform,
    message,
    device_types,
    options,
    actions,
    interactive,
    scheduled_time,
    local_scheduled_time,
    in_app,
    named_user,
]



from urbanairship import common
from datetime import datetime


class IndividualResponseStats(object):
    def __init__(self, airship):
        self.airship = airship

    def get(self, push_id):
        url = self.airship.urls.get('reports_url') + 'responses/' + push_id
        response = self.airship.request('GET', None, url, version=3)
        payload = response.json()
        return common.IteratorDataObj.from_payload(payload)


class ResponseList(common.IteratorParent):
    next_url = None
    data_attribute = 'pushes'

    def __init__(
            self, airship, start_date, end_date, limit=None, start_id=None
    ):
        if not airship or not start_date or not end_date:
            raise TypeError('airship, start_date, & end_date cannot be empty')
        if not isinstance(start_date, datetime) or not \
                isinstance(end_date, datetime):
            raise TypeError('start_date and end_date must be datetime objects')
        params = {
            'start': start_date.strftime('%Y-%m-%d %H:%M:%S'),
            'end': end_date.strftime('%Y-%m-%d %H:%M:%S'),
        }
        if limit:
            params['limit'] = limit
        if start_id:
            params['start_id'] = start_id
        self.next_url = airship.urls.get('reports_url') + 'responses/list'
        super(ResponseList, self).__init__(airship, params)


class DevicesReport(object):
    def __init__(self, airship):
        self.airship = airship

    def get(self, date):
        if not date:
            raise TypeError("date cannot be empty")
        if not isinstance(date, datetime):
            raise ValueError(
                'date must be a datetime object')
        url = self.airship.urls.get('reports_url') + 'devices/'
        params = {
            'date': date.strftime('%Y-%m-%dT%H:%M:%S')
        }
        response = self.airship._request(
            'GET', None, url, version=3, params=params
        )
        return response.json()


class ReportsList(common.IteratorParent):
    next_url = None
    data_attribute = None

    def __init__(self, airship, start_date, end_date, precision):
        if not airship or not start_date or not end_date or not precision:
            raise TypeError('None of the function parameters can be empty')

        if not isinstance(start_date, datetime) or not \
                isinstance(end_date, datetime):
            raise TypeError('start_date and end_date must be datetime objects')

        if precision not in ['HOURLY', 'DAILY', 'MONTHLY']:
            raise ValueError(
                "Precision must be 'HOURLY', 'DAILY', or 'MONTHLY'"
            )

        base_url = airship.urls.get('reports_url')

        if self.data_attribute == 'optins':
            self.next_url = base_url + 'optins/'
        elif self.data_attribute == 'optouts':
            self.next_url = base_url + 'optouts/'
        elif self.data_attribute == 'sends':
            self.next_url = base_url + 'sends/'
        elif self.data_attribute == 'responses':
            self.next_url = base_url + 'responses/'
        elif self.data_attribute == 'opens':
            self.next_url = base_url + 'opens/'
        elif self.data_attribute == 'timeinapp':
            self.next_url = base_url + 'timeinapp/'
        elif self.data_attribute == 'events':
            self.next_url = base_url + 'events/'

        params = {
            'start': start_date.strftime('%Y-%m-%d %H:%M:%S'),
            'end': end_date.strftime('%Y-%m-%d %H:%M:%S'),
            'precision': precision
        }

        super(ReportsList, self).__init__(airship, params)


class OptInList(ReportsList):
    data_attribute = 'optins'


class OptOutList(ReportsList):
    data_attribute = 'optouts'


class PushList(ReportsList):
    data_attribute = 'sends'


class ResponseReportList(ReportsList):
    data_attribute = 'responses'


class AppOpensList(ReportsList):
    data_attribute = 'opens'


class TimeInAppList(ReportsList):
    data_attribute = 'timeinapp'
    

class CustomEventsList(ReportsList):
    data_attribute = 'events'



import datetime
import json
import logging

from urbanairship import common


logger = logging.getLogger('urbanairship')


class Template(object):
    """Information object for a template.

    :ivar template_id: UUID; the ID of this template, set automatically.
    :ivar created_at: UTC datetime when the template was created.
    :ivar modified_at: UTC datetime when the template was last modified.
    :ivar last_used: UTC datetime when the template was last used for a push
        notification. Will be set to 'UNKNOWN' if never used.
    :ivar name: Required, name of the template, set during
        creation/modification.
    :ivar description: Optional, description of the template, set during
        creation/modification.
    :ivar variables: Required, an array of `Template Variable Objects
        <https://docs.urbanairship.com/api/ua/#template-variable-object>`_.
    :ivar push: Optional, a  push specification object, as defined by the `Push
        Object specification
        <https://docs.urbanairship.com/api/ua/#push-object>`_, but does not
        include ``audience`` or ``device_types`` since those are set when push
        messages are sent. Message Center ``message`` is also not supported in
        Templates at this time.

    """

    def __init__(
        self, airship, name=None, description=None, variables=None, push=None
    ):
        self.airship = airship
        self._template_id = None
        self._created_at = None
        self._modified_at = None
        self._last_used = None
        if name is not None:
            self.name = name
        else:
            self.name = None
        if description is not None:
            self.description = description
        else:
            self.description = None
        if variables is not None:
            self.variables = variables
        else:
            self.variables = []
        if push is not None:
            self.push = push
        else:
            self.push = {}

    @property
    def template_id(self):
        return self._template_id

    @property
    def created_at(self):
        return self._created_at

    @property
    def modified_at(self):
        return self._modified_at

    @property
    def last_used(self):
        return self._last_used

    @property
    def payload(self):
        data = {
            'name': self.name,
            'variables': self.variables,
            'push': self.push
        }
        if self.description is not None:
            data['description'] = self.description
        return data

    def create(self):
        """Create a notification template with the API.

        :raises AirshipFailure: Request failed.
        :raises Unauthorized: Authentication failed.

        """

        if not self.name:
            raise ValueError('Must set name before template creation.')

        if not self.push:
            raise ValueError('Must set push before template creation.')

        if 'message' in self.push.keys():
            raise ValueError(
                'Message center is not supported by templates.'
            )

        body = json.dumps(self.payload)
        response = self.airship._request(
            method='POST',
            body=body,
            url=self.airship.urls.get('templates_url'),
            content_type='application/json',
            version=3
        )
        self._template_id = response.json().get('template_id')
        logger.info(
            'Successful template creation for template %s', self.template_id
        )

        return response

    def update(self, template_id=None):
        """Update a template with the API.

        :raises AirshipFailure: Request failed.
        :raises Unauthorized: Authentication failed.

        """

        update_payload = {}

        if not self.name and not self.description \
                and not self.push and not self.variables:
            raise ValueError(
                'Must set at least one of name, description, push, or '
                'variables before template update.'
            )

        if self.name:
            update_payload['name'] = self.name

        if self.description:
            update_payload['description'] = self.description

        if 'message' in self.push.keys():
            raise ValueError(
                'Message center is not supported by templates.'
            )

        if self.push:
            update_payload['push'] = self.push

        if self.variables:
            update_payload['variables'] = self.variables

        if not template_id and not self.template_id:
            raise ValueError('Cannot update template without ID.')
        if template_id:
            self._template_id = template_id

        body = json.dumps(update_payload)
        response = self.airship._request(
            method='POST',
            body=body,
            url=self.airship.urls.get('templates_url') + self.template_id,
            content_type='application/json',
            version=3
        )
        logger.info(
            'Successful template update for template %s', self.template_id
        )

        return response

    def delete(self, template_id=None):
        """Delete a previously created template.

        :raises AirshipFailure: Request failed.
        :raises Unauthorized: Authentication failed.

        """

        if not template_id and not self.template_id:
            raise ValueError('Cannot delete template without ID.')
        if template_id:
            self._template_id = template_id

        response = self.airship._request(
            method='DELETE',
            body=None,
            url=self.airship.urls.get('templates_url') + self.template_id,
            version=3
        )
        logger.info(
            'Successful template delete for template %s', self.template_id
        )

        return response

    @classmethod
    def from_payload(cls, payload, id_key, airship):
        """Create based on results from a TemplateList iterator."""
        obj = cls(airship)
        obj._template_id = payload[id_key]
        for key in payload:
            if key in ('created_at', 'modified_at', 'last_used'):
                try:
                    payload[key] = datetime.datetime.strptime(
                        payload[key], '%Y-%m-%dT%H:%M:%S.%fZ'
                    )
                except:
                    payload[key] = 'UNKNOWN'
                setattr(obj, '_' + key, payload[key])
            elif key == 'template_id':
                obj._template_id = payload[key]
            else:
                setattr(obj, key, payload[key])
        return obj

    def lookup(self, template_id):
        """Fetch metadata from a template ID"""
        start_url = self.airship.urls.get('templates_url')
        data_attribute = 'template'
        id_key = 'id'
        params = {}
        url = start_url + template_id
        response = self.airship._request(
            method='GET',
            body=None,
            url=url,
            version=3,
            params=params
        )
        payload = response.json()
        return self.from_payload(payload[data_attribute], id_key, self.airship)


class TemplateList(common.IteratorParent):
    """Iterator for listing all templates for this application.

    :ivar limit: Number of entries to fetch in each page request.
    :returns: Each ``next`` returns a :py:class:`Template` object.

    """
    next_url = None
    data_attribute = 'templates'
    id_key = 'id'
    instance_class = Template

    def __init__(self, airship, limit=None):
        params = {'limit': limit} if limit else {}
        self.next_url = airship.urls.get('templates_url')
        super(TemplateList, self).__init__(airship, params)


def merge_data(template_id, substitutions):
    """Template push merge_data creation.

    :param template_id: Required, UUID.
    :param substitutions: Required, dictionary of template variables and their
        substitutions, e.g. {"FIRST_NAME": "Bob", "LAST_NAME": "Smith"}

    """

    md = {}

    md['template_id'] = template_id
    md['substitutions'] = {
        key: val for key, val in iter(substitutions.items()) if val is not None
    }

    return md




try:
    from setuptools import setup
except (ImportError):
    from distutils.core import setup

__about__ = {}

with open("urbanairship/__about__.py") as fp:
    exec(fp.read(), None, __about__)

setup(
    name="urbanairship",
    version=__about__["__version__"],
    author="Airship Customer Engineering",
    author_email="customer-engineering@airship.com",
    url="https://airship.com/",
    description="Python package for using the Airship API",
    long_description=open('README.rst').read(),
    packages=[
        "urbanairship",
        "urbanairship.push",
        "urbanairship.devices",
        "urbanairship.reports",
        "urbanairship.automation",
        "urbanairship.experiments"
    ],
    license='BSD License',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries'
    ],
    install_requires=[
        'requests>=1.2',
        'six'
    ],
)



import json
import logging
import re

logger = logging.getLogger('urbanairship')

VALID_MSISDN = re.compile(r'[0-9]*$')
VALID_SENDER = re.compile(r'[0-9]*$')


class Sms(object):
    """
    Create, register, opt-out and uninstall an Sms object.

    :param airship: Required. An urbanairship.Airship object instantiated with
        master authentication.
    :param sender: Required. The a number that recipients will recieve SMS
        notifications from. This must match your Urban Airship configuration.
    :param msisdn: Required. The mobile phone number you want to register as
        an SMS channel (or send a request to opt-in).
    :param opted_in: The UTC datetime in ISO 8601 format that represents the
        date and time when explicit permission was received from the user to
        receive messages. This is required for use with CreateAndSend.
    :param template_fields: For use with CreateAndSend with inline templates.
        A dict of template field names and their substitution values.
    """

    def __init__(self, airship, sender, msisdn, opted_in=False, template_fields=None):
        self.airship = airship
        self.sender = sender
        self.msisdn = msisdn
        self.opted_in = opted_in
        self.template_fields = template_fields
        self.channel_id = None

    @property
    def template_fields(self):
        return self._template_fields

    @template_fields.setter
    def template_fields(self, value):
        if not isinstance(value, (dict, type(None))):
            raise TypeError('template_fields must be a dict')

        self._template_fields = value

    @property
    def sender(self):
        return self._sender

    @sender.setter
    def sender(self, value):
        if not VALID_SENDER.match(value):
            raise ValueError('sender must be a numeric string')
        self._sender = value

    @property
    def msisdn(self):
        return self._msisdn

    @msisdn.setter
    def msisdn(self, value):
        if not VALID_MSISDN.match(value):
            raise ValueError('msisdn must be a numeric string')
        self._msisdn = value

    @property
    def common_payload(self):
        return {
            'sender': self.sender,
            'msisdn': self.msisdn,
        }

    @property
    def create_and_send_audience(self):
        audience = {
            'ua_sender': self.sender,
            'ua_msisdn': self.msisdn,
        }

        if self.template_fields:
            audience.update(self.template_fields)

        if self.opted_in:
            audience['ua_opted_in'] = self.opted_in
        else:
            raise ValueError(
                'sms objects for create and send must include opt-in datestamps'
                )
        return audience

    def register(self, opted_in=False):
        """Register an Sms channel with the sender ID and MSISDN

        :param opted_in: Optional UTC ISO 8601 datetime string that represents the
            date and time when explicit permission was received from the
            user to receive messages.

        :return: The response object from the api.
        """

        url = self.airship.urls.get('sms_url')
        reg_payload = self.common_payload

        if opted_in:
            reg_payload['opted_in'] = opted_in

        body = json.dumps(reg_payload).encode('utf-8')

        response = self.airship.request(
            method='POST',
            body=body,
            url=url,
            version=3
        )

        if response.json().get('status') == 'pending':
            logger.info(
                'Channel creation for msisdn %s pending user opt-in' % (
                    self.msisdn
                )
            )
        elif response.json().get('channel_id') is not None:
            self.channel_id = response.json().get('channel_id')
            logger.info(
                'Successfully registered Sms channel with channel_id %s' % (
                    self.channel_id
                )
            )
        else:
            logger.info(
                'Channel not yet created.'
            )

        return response

    def opt_out(self):
        """mark an sms channel at opted-out by sender ID and MSISDN

        :return: the response object from the api
        """

        url = self.airship.urls.get('sms_opt_out_url')

        response = self.airship.request(
            method='POST',
            body=json.dumps(self.common_payload).encode('utf-8'),
            url=url,
            version=3
        )

        logger.info(
            'Opted out Sms channel with sender: %s and msisdn: %s' % (
                self.sender, self.msisdn
            )
        )

        return response

    def uninstall(self):
        """Uninstall and remove all associated data from Urban Airship
        systems. Channel cannot be opted-in again. Use with caution.

        :return: the response object from the api"""

        url = self.airship.urls.get('sms_uninstall_url')

        response = self.airship.request(
            method='POST',
            body=json.dumps(self.common_payload).encode('utf-8'),
            url=url,
            version=3
        )

        logger.info(
            'Uninstalled Sms channel with sender: %s and msisdn: %s' % (
                self.sender, self.msisdn
            )
        )

        return response

    def lookup(self):
        """Look up Sms channel information

        :return: the response object from the api
        """

        url = self.airship.urls.get('sms_url') + '{msisdn}/{sender}'.format(
            msisdn=self.msisdn,
            sender=self.sender
        )

        response = self.airship.request(
            method='GET',
            body=None,
            url=url,
            version=3
        )

        return response



import json


class Automation(object):
    """An object for getting and creating automations"""

    def __init__(self, airship):
        self.airship = airship

    def create(self, pipelines):
        """Create an automation with one or more Pipeline payloads

        :keyword pipelines: A single Pipeline payload or list of Pipeline
        payloads
        """
        url = self.airship.urls.get('pipelines_url')
        body = json.dumps(pipelines)
        response = self.airship.request(
            method='POST',
            body=body,
            url=url,
            content_type='application/json',
            version=3
        )

        return response

    def validate(self, pipelines):
        """Validate a Pipeline payloads

        :keyword pipelines: A single Pipeline payload or list of Pipeline
        payloads
        """
        url = self.airship.urls.get('pipelines_url') + 'validate/'
        body = json.dumps(pipelines)
        response = self.airship.request(
            method='POST',
            body=body,
            url=url,
            content_type='application/json',
            version=3
        )

        return response

    def update(self, pipeline_id, pipeline):
        """Update an existing Automation Pipeline

        :keyword pipeline_id: A UA Pipeline ID
        :keyword pipeline: Full Pipeline payload; partial updates are not
        supported
        """
        url = self.airship.urls.get('pipelines_url') + pipeline_id
        body = json.dumps(pipeline)
        response = self.airship.request(
            method='PUT',
            body=body,
            url=url,
            content_type='application/json',
            version=3
        )

        return response

    def delete(self, pipeline_id):
        """Delete an existing Automation Pipeline

        :keyword pipeline_id: A UA Pipeline ID
        """
        url = self.airship.urls.get('pipelines_url') + pipeline_id
        response = self.airship.request(
            method='DELETE',
            body=None,
            url=url,
            version=3
        )

        return response

    def lookup(self, pipeline_id):
        """Lookup an Automation Pipeline

        :keyword pipeline_id: A UA Pipeline ID
        """
        url = self.airship.urls.get('pipelines_url') + pipeline_id
        response = self.airship.request(
            method='GET',
            body=None,
            url=url,
            version=3
        )

        return response

    def list_automations(self, limit=None, enabled=False):
        """List active Automations

        :keyword limit: Optional, maximum pipelines to return
        :keyword enabled: Optional, boolean limiter for results to only enabled
        Pipelines
        """
        params = {}
        if isinstance(limit, int):
            params['limit'] = limit
        if isinstance(enabled, bool):
            params['enabled'] = enabled

        url = self.airship.urls.get('pipelines_url')
        response = self.airship.request(
            method='GET',
            body=None,
            params=params,
            url=url,
            version=3
        )

        return response

    def list_deleted_automations(self, start=None):
        """List deleted Automation Pipelines

        :keyword start: Optional, starting timestamp for limiting results;
        ISO-8601 format
        """
        params = {}
        if start:
            params['start'] = start
        url = self.airship.urls.get('pipelines_url') + 'deleted/'
        response = self.airship.request(
            method='GET',
            body=None,
            params=params,
            url=url,
            version=3
        )

        return response



import json
import logging
import warnings

from urbanairship import devices


logger = logging.getLogger('urbanairship')


class Push(object):
    """A push notification. Set audience, message, etc, and send."""

    def __init__(self, airship):
        self._airship = airship
        self.audience = None
        self.notification = None
        self._device_types = None
        self.options = None
        self.campaigns = None
        self.message = None
        self.in_app = None

    @property
    def payload(self):
        data = {
            'audience': self.audience,
            'notification': self.notification,
            'device_types': self.device_types,
        }
        if self.options is not None:
            data['options'] = self.options
        if self.campaigns is not None:
            data['campaigns'] = self.campaigns
        if self.message is not None:
            data['message'] = self.message
        if self.in_app is not None:
            data['in_app'] = self.in_app
        return data

    @property
    def device_types(self):
        return self._device_types

    @device_types.setter
    def device_types(self, types):
        if types == 'all' or (len(types) == 1 and types[0] == 'all'):
            warnings.warn(
                "The device type 'all' has been deprecated.",
                DeprecationWarning
            )
        self._device_types = types

    def send(self):
        """Send the notification.

        :returns: :py:class:`PushResponse` object with ``push_ids`` and
            other response data.
        :raises AirshipFailure: Request failed.
        :raises Unauthorized: Authentication failed.
        :raises ValueError: Required keys missing or incorrect values included.
        """
        if 'email' in self.payload['notification']:
            if self.payload['device_types'] == 'all':
                raise ValueError(
                    'device_types cannot be all when including an email override'
                )
            if 'email' not in self.payload['device_types']:
                raise ValueError(
                    'email must be in device_types if email override is included'
                )
        if 'email' in self.payload['device_types'] \
                and 'email' not in self.payload['notification']:
            raise ValueError(
                'email override must be included when email is in device_types'
            )

        body = json.dumps(self.payload)
        response = self._airship._request(
            method='POST',
            body=body,
            url=self._airship.urls.get('push_url'),
            content_type='application/json',
            version=3
        )

        data = response.json()
        logger.info('Push successful. push_ids: %s',
                    ', '.join(data.get('push_ids', []))
                    )

        return PushResponse(response)


class ScheduledPush(object):
    """A scheduled push notification. Set schedule, push, and send."""

    def __init__(self, airship):
        self._airship = airship
        self.schedule = None
        self.name = None
        self.push = None
        self.url = None

    @classmethod
    def from_url(cls, airship, url):
        """Load an existing scheduled push from its URL."""

        sched = cls(airship)
        response = sched._airship._request(
            method='GET',
            body=None,
            url=url,
            version=3
        )
        payload = response.json()
        sched.name = payload.get('name')
        sched.schedule = payload['schedule']
        sched.push = Push(airship)
        sched.push.audience = payload['push']['audience']
        sched.push.notification = payload['push']['notification']
        sched.push.device_types = payload['push']['device_types']
        if 'message' in payload['push']:
            sched.push.message = payload['push']['message']
        if 'options' in payload['push']:
            sched.push.options = payload['push']['options']
        sched.url = url
        return sched

    @classmethod
    def from_payload(cls, payload, id_key, airship):
        """Create based on results from a ScheduledList iterator."""
        obj = cls(airship)
        obj._schedule_id = payload[id_key]
        for key in payload:
            setattr(obj, key, payload[key])
        return obj

    @property
    def payload(self):
        if hasattr(self.push, 'merge_data'):  # create template payload
            data = self.push.payload
            data['schedule'] = self.schedule
        elif isinstance(self.push, CreateAndSendPush):  # create cas payload
            if 'scheduled_time' not in self.schedule:
                raise ValueError(
                    'only scheduled_time supported with create and send schedules'
                )
            data = {
                'schedule': self.schedule,
                'push': self.push.payload
            }
        else:
            data = {
                'schedule': self.schedule,
                'push': self.push.payload,
            }

        if self.name is not None:
            data['name'] = self.name

        return data

    def send(self):
        """Schedule the notification

        :returns: :py:class:`PushResponse` object with ``schedule_url`` and
            other response data.
        :raises AirshipFailure: Request failed.
        :raises Unauthorized: Authentication failed.

        """
        body = json.dumps(self.payload)

        if hasattr(self.push, 'merge_data'):
            url = self._airship.urls.get('schedule_template_url')
        elif isinstance(self.push, CreateAndSendPush):
            url = self._airship.urls.get('schedule_create_and_send_url')
        else:
            url = self._airship.urls.get('schedules_url')

        response = self._airship._request(
            method='POST',
            body=body,
            url=url,
            content_type='application/json',
            version=3
        )
        data = response.json()

        urls = data.get('schedule_urls', [])
        if urls:
            self.url = urls[0]
            logger.info('Scheduled push successful. schedule_urls: %s',
                        ', '.join(data.get('schedule_urls', [])))

        else:
            logger.info('Scheduled push resulted in zero messages scheduled.')

        return PushResponse(response)

    def cancel(self):
        """Cancel a previously scheduled notification."""
        if not self.url:
            raise ValueError('Cannot cancel ScheduledPush without url.')

        self._airship._request(
            method='DELETE',
            body=None,
            url=self.url,
            version=3
        )

    def update(self):
        if not self.url:
            raise ValueError(
                'Cannot update ScheduledPush without url.')
        body = json.dumps(self.payload)
        response = self._airship._request(
            method='PUT',
            body=body,
            url=self.url,
            content_type='application/json',
            version=3
        )

        data = response.json()
        logger.info('Scheduled push update successful. schedule_urls: %s',
                    ', '.join(data.get('schedule_urls', [])))

        return PushResponse(response)


class TemplatePush(object):
    """A personalized push notification. Set details and send."""

    def __init__(self, airship):
        self._airship = airship
        self.audience = None
        self.device_types = None
        self.merge_data = None

    @property
    def payload(self):
        data = {
            'audience': self.audience,
            'device_types': self.device_types,
            'merge_data': self.merge_data
        }

        return data

    def send(self):
        """Send the personalized notification.

        :returns: :py:class:`PushResponse` object with ``push_ids`` and
            other response data.
        :raises AirshipFailure: Request failed.
        :raises Unauthorized: Authentication failed.

        """

        if not self.audience:
            raise ValueError('Must set audience for template push.')

        if not self.device_types:
            raise ValueError('Must set device_types for template push.')

        body = json.dumps(self.payload)
        response = self._airship._request(
            method='POST',
            body=body,
            url=self._airship.urls.get('templates_url') + 'push',
            content_type='application/json',
            version=3
        )

        data = response.json()
        logger.info('Push successful. push_ids: %s',
                    ', '.join(data.get('push_ids', []))
                    )

        return PushResponse(response)


class CreateAndSendPush(object):
    """
    Creates and sends to email, sms or open channels. Channel ids are created
    but not returned by this request. Use lookup/listing endpoints to find
    channel_id values. Opt-in date attributes are required on Sms and Email
    objects passed in.

    :param airship: Required. An urbanairship.Airship object instantiated with
        master authentication.
    :param channels: Required. A list of Sms, Email or OpenChannel objects.
        channels may only be of one type and must match the single value for
        CreateAndSend.device_types.
    """

    def __init__(self, airship, channels=[]):
        self._airship = airship
        self.channels = channels
        self.notification = None
        self.campaigns = None

    @property
    def device_types(self):
        return self._device_types

    @device_types.setter
    def device_types(self, values):
        accepted_device_types = ('sms', 'email', 'open::')

        if len(values) != 1:
            raise ValueError('only a single device_type may be used.')

        for value in values:
            if value[:6] not in accepted_device_types:
                raise ValueError(
                    'device_types must be one of {}'.format(
                        str(accepted_device_types)
                        )
                    )

        self._device_types = values

    @property
    def audience(self):
        if 'email' in self.device_types:
            return self._email_audience()
        elif 'sms' in self.device_types:
            return self._sms_audience()
        else:
            return self._open_channel_audience()

    @property
    def channels(self):
        return self._channels

    @channels.setter
    def channels(self, value):
        if type(value) is not list:
            raise TypeError('channels must be a list')
        if len(value) > 1000:
            raise ValueError('channels list must have 1000 or fewer items')

        self._channels = value

    @property
    def payload(self):
        data = {
            'audience': self.audience,
            'notification': self.notification,
            'device_types': self.device_types
        }
        if self.campaigns is not None:
            data['campaigns'] = self.campaigns
        return data

    def _email_audience(self):
        addresses = []
        for email in self.channels:
            if not isinstance(email, devices.Email):
                raise TypeError(
                    'Can only use email channels when device_types is email'
                )
            addresses.append(email.create_and_send_audience)
        audience = {'create_and_send': addresses}

        return audience

    def _sms_audience(self):
        addresses = []
        for sms in self.channels:
            if not isinstance(sms, devices.Sms):
                raise TypeError(
                    'Can only use Sms objects when device_types is sms'
                    )
            addresses.append(sms.create_and_send_audience)
        audience = {'create_and_send': addresses}

        return audience

    def _open_channel_audience(self):
        addresses = []
        for open_channel in self.channels:
            if not isinstance(open_channel, devices.OpenChannel):
                raise TypeError(
                    'Can only use OpenChannel objects when device_types is open::'
                )
            addresses.append(open_channel.create_and_send_audience)
        audience = {'create_and_send': addresses}

        return audience

    def send(self):
        """Send the notification.

        :returns: :py:class:`PushResponse` object with ``push_ids`` and
            other response data.
        :raises AirshipFailure: Request failed.
        :raises Unauthorized: Authentication failed.
        :raises ValueError: Required keys missing or incorrect values included.
        """
        body = json.dumps(self.payload)
        response = self._airship._request(
            method='POST',
            body=body,
            url=self._airship.urls.get('create_and_send_url'),
            content_type='application/json',
            version=3
        )

        logger.info('Create and Send successful')

        return PushResponse(response)


class PushResponse(object):
    """Response to a successful push notification send or schedule.

    Right now this is a fairly simple wrapper around the json payload response,
    but making it an object gives us some flexibility to add functionality
    later.

    """
    ok = None
    push_ids = None
    schedule_url = None
    operation_id = None
    payload = None

    def __init__(self, response):
        data = response.json()
        self.push_ids = data.get('push_ids')
        self.schedule_url = data.get('schedule_urls', [])
        self.operation_id = data.get('operation_id')
        self.ok = data.get('ok')
        self.payload = data

    def __str__(self):
        return 'Response Payload: {0}'.format(self.payload)



class Experiment(object):
    """ An experiment object describes an A/B test,
    including the audience and variant portions.
    """

    def __init__(self,
                 audience,
                 device_types,
                 variants,
                 name=None,
                 description=None,
                 weight=None,
                 campaigns=None,
                 control=None
                 ):
        """
        :keyword audience: [required] The audience for the experiment
        :keyword device_types: An array containing one or more strings identifying
            targeted platforms. Accepted platforms are ios, android, amazon, wns, web,
            sms, email, and open::<open_platform_name>
        :keywors variants: [required] The variants for the experiment. An experiment
            must have at least 1 variant and no more than 26.
        :keyword name: [optional] A name for the experiment
        :keyword description: [optional] A description of the experiment
        :keyword campaigns: [optional] Campaigns object that will be applied to
            resulting pushes
        :keyword control: [optional] The proportional subset of the audience that will
             not receive a push

        """
        self.audience = audience
        self.device_types = device_types
        self.variants = variants
        self.name = name
        self.description = description
        self.campaigns = campaigns
        self.control = control
        self.weight = weight

    @property
    def payload(self):
        """JSON serialized experiment object"""

        variants_data = []
        for variant in self.variants:
            variant_data = {}
            push_options = {}

            if getattr(variant, 'description', None):
                variant_data['description'] = variant.description
            if getattr(variant, 'name', None):
                variant_data['name'] = variant.name

            if getattr(variant.push, 'in_app', None):
                push_options['in_app'] = variant.push.in_app
            if getattr(variant.push, 'notification', None):
                push_options['notification'] = variant.push.notification
            if getattr(variant.push, 'options', None):
                push_options['options'] = variant.push.options
            if getattr(variant, 'schedule', None):
                variant_data['schedule'] = variant.schedule
            if getattr(variant, 'weight', None):
                variant_data['weight'] = variant.weight

            variant_data['push'] = push_options
            variants_data.append(variant_data)

        data = {
            "audience": self.audience,
            "device_types": self.device_types,
            "variants": variants_data
        }

        if self.name is not None:
            data["name"] = self.name
        if self.description is not None:
            data["description"] = self.description
        if self.campaigns is not None:
            data["campaigns"] = self.campaigns
        if self.control is not None:
            data["control"] = self.control

        if self.weight is not None:
            variant_data['weight'] = self.weight

        return data

    @property
    def name(self):
        if not self._name:
            return None
        return self._name

    @name.setter
    def name(self, value):
        if not isinstance(value, str):
            TypeError(
                'the name must be a string type'
            )

        self._name = value

    @property
    def description(self):
        if not self._description:
            return None
        return self._description

    @description.setter
    def description(self, value):
        if not isinstance(value, str):
            TypeError(
                'the description must be type string'
            )

        self._description = value

    @property
    def control(self):
        if not self._control:
            return None
        return self._control

    @control.setter
    def control(self, value):
        if not isinstance(value, float):
            TypeError(
                'the control must be type float'
            )
        if value is not None:
            if not 0.0 >= value >= 1.0:
                ValueError(
                    'control must be in a range of 0.0 and 1.0'
                )

        self._control = value



import json
import logging

logger = logging.getLogger('urbanairship')


class ChannelTags(object):
    """Modify the tags for a channel"""

    def __init__(self, airship):
        self.url = airship.urls.get('channel_url') + 'tags/'
        self._airship = airship
        self.audience = {}
        self.add_group = {}
        self.remove_group = {}
        self.set_group = {}

    def set_audience(self, ios=None, android=None, amazon=None, web=None):
        if ios is not None:
            self.audience['ios_channel'] = ios
        if android is not None:
            self.audience['android_channel'] = android
        if amazon is not None:
            self.audience['amazon_channel'] = amazon
        if web is not None:
            self.audience['channel'] = web

    def add(self, group_name, tags):
        self.add_group[group_name] = tags

    def remove(self, group_name, tags):
        self.remove_group[group_name] = tags

    def set(self, group_name, tags):
        self.set_group[group_name] = tags

    def send(self):
        payload = {}

        if not self.audience:
            raise ValueError('A audience is required for modifying tags')
        payload['audience'] = self.audience

        if self.add_group:
            if self.set_group:
                raise ValueError('A tag request cannot contain both an "add"'
                                 ' and a "set" field.')
            payload['add'] = self.add_group

        if self.remove_group:
            if self.set_group:
                raise ValueError('A tag request cannot contain both a "remove"'
                                 ' and a "set" field.')
            payload['remove'] = self.remove_group

        if self.set_group:
            payload['set'] = self.set_group

        if not self.add_group and not self.remove_group and not self.set_group:
            raise ValueError('An add, remove, or set field was not set')

        body = json.dumps(payload)
        response = self._airship._request(
            'POST', body, self.url,
            'application/json', version=3
        )
        return response.json()


class OpenChannelTags(object):
    """Modify the tags for an open channel"""

    def __init__(self, airship):
        self.url = airship.urls.get('open_channel_url') + 'tags/'
        self._airship = airship
        self.audience = {}
        self.add_group = {}
        self.remove_group = {}
        self.set_group = {}

    def set_audience(self, address, open_platform):
        self.audience = {
            'address': address,
            'open_platform_name': open_platform
        }

    def add(self, group_name, tags):
        self.add_group[group_name] = tags

    def remove(self, group_name, tags):
        self.remove_group[group_name] = tags

    def set(self, group_name, tags):
        self.set_group[group_name] = tags

    def send(self):
        payload = {}

        if not self.audience:
            raise ValueError('An audience is required to modify tags')
        payload['audience'] = self.audience

        if not self.add_group and not self.remove_group and not self.set_group:
            raise ValueError('An add, remove, or set field was not set')

        if self.set_group:
            if self.add_group or self.remove_group:
                raise ValueError('A "set" tag request cannot contain '
                                 '"add" or "remove" fields')
            payload['set'] = self.set_group

        if self.add_group:
            payload['add'] = self.add_group

        if self.remove_group:
            payload['remove'] = self.remove_group

        body = json.dumps(payload)
        response = self._airship._request(
            'POST', body, self.url,
            'application/json', version=3
        )
        return response.json()



import json


class ABTest(object):
    def __init__(self, airship):
        self.airship = airship

    def _get_listing(self, url, limit=None):
        """List experiments

        :keyword limit: Optional, maximum number of experiments,
        default is 10, max is 100
        """

        params = {}
        if isinstance(limit, int):
            params['limit'] = limit

        response = self.airship.request(
            method='GET',
            body=None,
            params=params,
            url=url,
            version=3
        )
        return response

    def list_experiments(self):
        """ List experiments, sorted by created_at date/time from newest to oldest 

        :keyword limit: Positive maximum number of elements to return per page.
        Default limit is 10. Max: 100 and Min: 1.
        """
        url = self.airship.urls.get('experiments_url')
        return self._get_listing(url)

    def create(self, experiment):
        """ Create an experiment """

        url = self.airship.urls.get('experiments_url')
        body = json.dumps(experiment.payload)
        response = self.airship.request(
            method='POST',
            body=body,
            url=url,
            content_type='application/json',
            version=3
        )
        return response

    def list_scheduled_experiment(self):
        """ List scheduled experiments in order, from closest to the current
        date-time to farthest """

        url = self.airship.urls.get('experiments_schedule_url')
        return self._get_listing(url)

    def delete(self, experiment_id):
        """ Delete a scheduled experiment. You can only delete experiments before they start

        :keyword experiment_id: The unique identifier of the experiment, type string
        DELETE /api/experiments/scheduled/{experiment_id}
        """

        url = self.airship.urls.get('experiments_schedule_url') + '/' + experiment_id
        response = self.airship.request(
            method='DELETE',
            body=None,
            url=url,
            version=3
        )

        return response

    def validate(self, experiment):
        """ Accepts the same range of payloads as /api/experiments,
        but only parses and validates the payload without creating the experiment.
        An experiment may validate and still fail to be delivered. For example,
        you may have a valid experiment with no devices in your audience.
            POST /api/experiments/validate

        :keyword experiment: Body of the experiment you want to validate
        """
        url = self.airship.urls.get('experiments_validate')
        body = json.dumps(experiment.payload)
        response = self.airship.request(
            method='POST',
            body=body,
            url=url,
            content_type='application/json',
            version=3
        )

        return response

    def lookup(self, experiment_id):
        """ Look up an experiment (A/B Test)

        :keyword experiment_id: The unique identifier of the experiment, type string
        """

        url = self.airship.urls.get('experiments_url') + '/' + experiment_id
        response = self.airship.request(
            method='GET',
            body=None,
            url=url,
            version=3
        )

        return response



import logging
import re

import requests

from . import common, __about__
from .push import Push, ScheduledPush, TemplatePush

logger = logging.getLogger('urbanairship')

VALID_KEY = re.compile(r'^[\w-]{22}$')
VALID_LOCATIONS = ['eu', 'us', None]


class Urls(object):
    def __init__(self, location=None):
        if not location or location.lower() == 'us':
            self.base_url = 'https://go.urbanairship.com/api/'
        elif location.lower() == 'eu':
            self.base_url = 'https://go.airship.eu/api/'

        self.channel_url = self.base_url + 'channels/'
        self.open_channel_url = self.channel_url + 'open/'
        self.device_token_url = self.base_url + 'device_tokens/'
        self.apid_url = self.base_url + 'apids/'
        self.push_url = self.base_url + 'push/'
        self.schedules_url = self.base_url + 'schedules/'
        self.tags_url = self.base_url + 'tags/'
        self.segments_url = self.base_url + 'segments/'
        self.reports_url = self.base_url + 'reports/'
        self.lists_url = self.base_url + 'lists/'
        self.location_url = self.base_url + 'location/'

        self.templates_url = self.base_url + 'templates/'
        self.schedule_template_url = self.templates_url + 'schedules/'

        self.pipelines_url = self.base_url + 'pipelines/'

        self.named_user_url = self.base_url + 'named_users/'
        self.named_user_tag_url = self.named_user_url + 'tags/'
        self.named_user_disassociate_url = self.named_user_url + 'disassociate/'
        self.named_user_associate_url = self.named_user_url + 'associate/'

        self.sms_url = self.channel_url + 'sms/'
        self.sms_opt_out_url = self.sms_url + 'opt-out/'
        self.sms_uninstall_url = self.sms_url + 'uninstall/'

        self.email_url = self.channel_url + 'email/'
        self.email_tags_url = self.email_url + 'tags/'
        self.email_uninstall_url = self.email_url + 'uninstall/'

        self.create_and_send_url = self.base_url + 'create-and-send/'
        self.schedule_create_and_send_url = self.schedules_url + 'create-and-send/'

        self.experiments_url = self.base_url + 'experiments/'
        self.experiments_schedule_url = self.experiments_url + 'scheduled/'
        self.experiments_validate = self.experiments_url + 'validate/'

    def get(self, endpoint):
        url = getattr(self, endpoint, None)

        if not url:
            raise AttributeError('No url for endpoint %s' % endpoint)

        return url


class Airship(object):

    def __init__(self, key, secret, location=None):
        self.key = key
        self.secret = secret
        self.location = location

        self.session = requests.Session()
        self.session.auth = (key, secret)
        self.urls = Urls(location)

    @property
    def key(self):
        return self._key

    @key.setter
    def key(self, value):
        if not VALID_KEY.match(value):
            raise ValueError('keys must be 22 characters')
        self._key = value

    @property
    def location(self):
        return self._location

    @location.setter
    def location(self, value):
        if value not in VALID_LOCATIONS:
            raise ValueError('location must be {} or None'.format(VALID_LOCATIONS))
        self._location = value

    @property
    def secret(self):
        return self._secret

    @secret.setter
    def secret(self, value):
        if not VALID_KEY.match(value):
            raise ValueError('secrets must be 22 characters')
        self._secret = value

    def request(self, method, body, url,
                content_type=None, version=None, params=None):
        return self._request(method, body, url,
                             content_type, version, params)

    def _request(self, method, body, url, content_type=None,
                 version=None, params=None, encoding=None):

        headers = \
                {'User-agent': 'UAPythonLib/{0}'.format(__about__.__version__)}
        if content_type:
            headers['Content-type'] = content_type
        if version:
            headers['Accept'] = ('application/vnd.urbanairship+json; '
                                 'version=%d;' % version)
        if encoding:
            headers['Content-Encoding'] = encoding

        logger.debug(
            'Making %s request to %s. Headers:\n\t%s\nBody:\n\t%s',
            method,
            url,
            '\n\t'.join(
                '%s: %s' % (key, value) for (key, value) in headers.items()
            ),
            body
        )

        response = self.session.request(
            method, url, data=body, params=params, headers=headers)

        logger.debug(
            'Received %s response. Headers:\n\t%s\nBody:\n\t%s',
            response.status_code,
            '\n\t'.join(
                '%s: %s' % (key, value) for (key, value)
                in response.headers.items()
            ),
            response.content
        )

        if response.status_code == 401:
            raise common.Unauthorized
        elif not (200 <= response.status_code < 300):
            raise common.AirshipFailure.from_response(response)

        return response

    def create_push(self):
        """Create a Push notification."""
        return Push(self)

    def create_scheduled_push(self):
        """Create a Scheduled Push notification."""
        return ScheduledPush(self)

    def create_template_push(self):
        """Create a Scheduled Push notification."""
        return TemplatePush(self)
